import asyncio
from unittest.mock import Mock

import kopf
from kopf.reactor.indexing import OperatorIndexers
from kopf.reactor.processing import process_resource_event
from kopf.structs.containers import ResourceMemories
from kopf.structs.ephemera import Memo


async def test_parameter_is_passed_when_specified(resource, cause_mock, registry, settings):
    mock = Mock()

    # If it works for this handler, we assume it works for all of them.
    # Otherwise, it is too difficult to trigger the actual invocation.
    @kopf.on.event(*resource, param=123)
    def fn(**kwargs):
        mock(**kwargs)

    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': None, 'object': {}},
        event_queue=event_queue,
    )

    assert mock.called
    assert mock.call_args_list[0][1]['param'] == 123


async def test_parameter_is_passed_even_if_not_specified(resource, cause_mock, registry, settings):
    mock = Mock()

    # If it works for this handler, we assume it works for all of them.
    # Otherwise, it is too difficult to trigger the actual invocation.
    @kopf.on.event(*resource)
    def fn(**kwargs):
        mock(**kwargs)

    event_queue = asyncio.Queue()
    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': None, 'object': {}},
        event_queue=event_queue,
    )

    assert mock.called
    assert mock.call_args_list[0][1]['param'] is None



import asyncio

import pytest

from kopf.structs.primitives import Toggle, ToggleSet


@pytest.mark.parametrize('fn, expected', [(all, True), (any, False)])
async def test_created_empty(fn, expected):
    toggleset = ToggleSet(fn)
    assert len(toggleset) == 0
    assert set(toggleset) == set()
    assert Toggle() not in toggleset
    assert toggleset.is_on() == expected
    assert toggleset.is_off() == (not expected)


@pytest.mark.parametrize('fn', [all, any])
async def test_making_a_default_toggle(fn):
    toggleset = ToggleSet(fn)
    toggle = await toggleset.make_toggle()
    assert len(toggleset) == 1
    assert set(toggleset) == {toggle}
    assert toggle in toggleset
    assert Toggle() not in toggleset
    assert toggleset.is_on() == False
    assert toggleset.is_off() == True


@pytest.mark.parametrize('fn', [all, any])
async def test_making_a_turned_off_toggle(fn):
    toggleset = ToggleSet(fn)
    toggle = await toggleset.make_toggle(False)
    assert len(toggleset) == 1
    assert set(toggleset) == {toggle}
    assert toggle in toggleset
    assert Toggle() not in toggleset
    assert toggleset.is_on() == False
    assert toggleset.is_off() == True


@pytest.mark.parametrize('fn', [all, any])
async def test_making_a_turned_on_toggle(fn):
    toggleset = ToggleSet(fn)
    toggle = await toggleset.make_toggle(True)
    assert len(toggleset) == 1
    assert set(toggleset) == {toggle}
    assert toggle in toggleset
    assert Toggle() not in toggleset
    assert toggleset.is_on() == True
    assert toggleset.is_off() == False


@pytest.mark.parametrize('fn, expected', [(all, True), (any, False)])
async def test_dropping_a_turned_off_toggle(fn, expected):
    toggleset = ToggleSet(fn)
    toggle = await toggleset.make_toggle(False)
    await toggle.turn_to(True)
    await toggleset.drop_toggle(toggle)
    assert len(toggleset) == 0
    assert set(toggleset) == set()
    assert toggle not in toggleset
    assert toggleset.is_on() == expected
    assert toggleset.is_off() == (not expected)


@pytest.mark.parametrize('fn, expected', [(all, True), (any, False)])
async def test_dropping_a_turned_on_toggle(fn, expected):
    toggleset = ToggleSet(fn)
    toggle = await toggleset.make_toggle(True)
    await toggleset.drop_toggle(toggle)
    assert len(toggleset) == 0
    assert set(toggleset) == set()
    assert toggle not in toggleset
    assert toggleset.is_on() == expected
    assert toggleset.is_off() == (not expected)


@pytest.mark.parametrize('fn, expected', [(all, True), (any, False)])
async def test_dropping_an_unexistent_toggle(fn, expected):
    toggleset = ToggleSet(fn)
    toggle = Toggle()
    await toggleset.drop_toggle(toggle)
    assert len(toggleset) == 0
    assert set(toggleset) == set()
    assert toggle not in toggleset
    assert toggleset.is_on() == expected
    assert toggleset.is_off() == (not expected)


@pytest.mark.parametrize('fn, expected', [(all, True), (any, False)])
async def test_dropping_multiple_toggles(fn, expected):
    toggleset = ToggleSet(fn)
    toggle1 = await toggleset.make_toggle(True)
    toggle2 = Toggle()
    await toggleset.drop_toggles([toggle1, toggle2])
    assert len(toggleset) == 0
    assert set(toggleset) == set()
    assert toggle1 not in toggleset
    assert toggle2 not in toggleset
    assert toggleset.is_on() == expected
    assert toggleset.is_off() == (not expected)


@pytest.mark.parametrize('fn', [all, any])
async def test_turning_a_toggle_on_turns_the_toggleset_on(fn):
    toggleset = ToggleSet(fn)
    toggle = await toggleset.make_toggle(False)
    assert toggleset.is_on() == False
    assert toggleset.is_off() == True

    await toggle.turn_to(True)
    assert toggleset.is_on() == True
    assert toggleset.is_off() == False


@pytest.mark.parametrize('fn', [all, any])
async def test_turning_a_toggle_off_turns_the_toggleset_off(fn):
    toggleset = ToggleSet(fn)
    toggle = await toggleset.make_toggle(True)
    assert toggleset.is_on() == True
    assert toggleset.is_off() == False

    await toggle.turn_to(False)
    assert toggleset.is_on() == False
    assert toggleset.is_off() == True


@pytest.mark.parametrize('fn', [all])
async def test_all_toggles_must_be_on_for_alltoggleset_to_be_on(fn):
    toggleset = ToggleSet(fn)
    toggle1 = await toggleset.make_toggle(False)
    toggle2 = await toggleset.make_toggle(False)
    assert toggleset.is_on() == False
    assert toggleset.is_off() == True

    await toggle1.turn_to(True)
    assert toggleset.is_on() == False
    assert toggleset.is_off() == True

    await toggle2.turn_to(True)
    assert toggleset.is_on() == True
    assert toggleset.is_off() == False


@pytest.mark.parametrize('fn', [all])
async def test_any_toggle_must_be_off_for_alltoggleset_to_be_off(fn):
    toggleset = ToggleSet(fn)
    toggle1 = await toggleset.make_toggle(True)
    toggle2 = await toggleset.make_toggle(True)
    assert toggleset.is_on() == True
    assert toggleset.is_off() == False

    await toggle1.turn_to(False)
    assert toggleset.is_on() == False
    assert toggleset.is_off() == True

    await toggle2.turn_to(False)
    assert toggleset.is_on() == False
    assert toggleset.is_off() == True


@pytest.mark.parametrize('fn', [any])
async def test_any_toggle_must_be_on_for_anytoggleset_to_be_on(fn):
    toggleset = ToggleSet(fn)
    toggle1 = await toggleset.make_toggle(False)
    toggle2 = await toggleset.make_toggle(False)
    assert toggleset.is_on() == False
    assert toggleset.is_off() == True

    await toggle1.turn_to(True)
    assert toggleset.is_on() == True
    assert toggleset.is_off() == False

    await toggle2.turn_to(True)
    assert toggleset.is_on() == True
    assert toggleset.is_off() == False


@pytest.mark.parametrize('fn', [any])
async def test_all_toggles_must_be_off_for_anytoggleset_to_be_off(fn):
    toggleset = ToggleSet(fn)
    toggle1 = await toggleset.make_toggle(True)
    toggle2 = await toggleset.make_toggle(True)
    assert toggleset.is_on() == True
    assert toggleset.is_off() == False

    await toggle1.turn_to(False)
    assert toggleset.is_on() == True
    assert toggleset.is_off() == False

    await toggle2.turn_to(False)
    assert toggleset.is_on() == False
    assert toggleset.is_off() == True


@pytest.mark.parametrize('fn', [all, any])
async def test_waiting_until_on_fails_when_not_turned_on(fn):
    toggleset = ToggleSet(fn)
    await toggleset.make_toggle(False)
    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(toggleset.wait_for(True), timeout=0.1)
    assert toggleset.is_off()


@pytest.mark.parametrize('fn', [all, any])
async def test_waiting_until_off_fails_when_not_turned_off(fn):
    toggleset = ToggleSet(fn)
    await toggleset.make_toggle(True)
    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(toggleset.wait_for(False), timeout=0.1)
    assert toggleset.is_on()


@pytest.mark.parametrize('fn', [all, any])
async def test_waiting_until_on_wakes_when_turned_on(fn, timer):
    toggleset = ToggleSet(fn)
    toggle = await toggleset.make_toggle(False)

    async def delayed_turning_on(delay: float):
        await asyncio.sleep(delay)
        await toggle.turn_to(True)

    with timer:
        asyncio.create_task(delayed_turning_on(0.05))
        await asyncio.wait_for(toggleset.wait_for(True), timeout=1.0)

    assert toggleset.is_on()
    assert timer.seconds < 0.5  # approx. 0.05 plus some code overhead


@pytest.mark.parametrize('fn', [all, any])
async def test_waiting_until_off_wakes_when_turned_off(fn, timer):
    toggleset = ToggleSet(fn)
    toggle = await toggleset.make_toggle(True)

    async def delayed_turning_off(delay: float):
        await asyncio.sleep(delay)
        await toggle.turn_to(False)

    with timer:
        asyncio.create_task(delayed_turning_off(0.05))
        await asyncio.wait_for(toggleset.wait_for(False), timeout=1.0)

    assert toggleset.is_off()
    assert timer.seconds < 0.5  # approx. 0.05 plus some code overhead


@pytest.mark.parametrize('fn', [all, any])
async def test_secures_against_usage_as_a_boolean(fn):
    toggle = ToggleSet(fn)
    with pytest.raises(NotImplementedError):
        bool(toggle)


@pytest.mark.parametrize('fn', [all, any])
async def test_repr_when_empty(fn):
    toggleset = ToggleSet(fn)
    assert repr(toggleset) == "set()"


@pytest.mark.parametrize('fn', [all, any])
async def test_repr_when_unnamed_and_off(fn):
    toggleset = ToggleSet(fn)
    await toggleset.make_toggle(False)
    assert repr(toggleset) == "{<Toggle: off>}"


@pytest.mark.parametrize('fn', [all, any])
async def test_repr_when_unnamed_and_on(fn):
    toggleset = ToggleSet(fn)
    await toggleset.make_toggle(True)
    assert repr(toggleset) == "{<Toggle: on>}"


@pytest.mark.parametrize('fn', [all, any])
async def test_repr_when_named_and_off(fn):
    toggleset = ToggleSet(fn)
    await toggleset.make_toggle(False, name='xyz')
    assert repr(toggleset) == "{<Toggle: xyz: off>}"


@pytest.mark.parametrize('fn', [all, any])
async def test_repr_when_named_and_on(fn):
    toggleset = ToggleSet(fn)
    await toggleset.make_toggle(True, name='xyz')
    assert repr(toggleset) == "{<Toggle: xyz: on>}"



import json

import pytest

from kopf.storage.conventions import StorageKeyFormingConvention
from kopf.storage.diffbase import AnnotationsDiffBaseStorage
from kopf.storage.progress import AnnotationsProgressStorage, ProgressRecord, SmartProgressStorage
from kopf.structs.bodies import Body
from kopf.structs.handlers import HandlerId
from kopf.structs.patches import Patch

ANNOTATIONS_POPULATING_STORAGES = [
    AnnotationsProgressStorage,
    SmartProgressStorage,
]
STORAGE_KEY_FORMING_CLASSES = [
    StorageKeyFormingConvention,
    AnnotationsProgressStorage,
    AnnotationsDiffBaseStorage,
]

CONTENT_DATA = ProgressRecord(
    started='2020-01-01T00:00:00',
    stopped='2020-12-31T23:59:59',
    delayed='3000-01-01T00:00:00',
    retries=0,
    success=False,
    failure=False,
    message=None,
)

CONTENT_JSON = json.dumps(CONTENT_DATA, separators=(',', ':'))

COMMON_KEYS = [
    # For character replacements (only those that happen in our own ids, not all of them).
    ['my-operator.example.com', 'a_b.c-d/e', 'my-operator.example.com/a_b.c-d.e'],
]

V1_KEYS = [
    # For V1 length cutting. Hint: the prefix length is 23, the remaining space is 63 - 23 - 1 = 39.
    # The suffix itself (if appended) takes 9, so it is 30 left. The same math for no prefix.
    ['my-operator.example.com', 'x', 'my-operator.example.com/x'],
    ['my-operator.example.com', 'x' * 39, 'my-operator.example.com/' + 'x' * 39],
    ['my-operator.example.com', 'x' * 40, 'my-operator.example.com/' + 'x' * 30 + 'xx-tEokcg'],
    ['my-operator.example.com', 'y' * 40, 'my-operator.example.com/' + 'y' * 30 + 'yy-VZlvhw'],
    ['my-operator.example.com', 'z' * 40, 'my-operator.example.com/' + 'z' * 30 + 'zz-LlPQyA'],

    # For special chars in base64 encoding ("+" and "/"), which are not compatible with K8s.
    # The numbers are found empirically so that both "/" and "+" are found in the base64'ed digest.
    ['my-operator.example.com', 'fn' * 323, 'my-operator.example.com/' + 'fn' * 15 + 'fn-Az-r.g'],
]

V2_KEYS = [
    # For V2 length cutting: 63 for the name part only, not the whole annotation.
    # The suffix itself (if appended) takes 9, so it is 63-9=54 left.
    ['my-operator.example.com', 'x', 'my-operator.example.com/x'],
    ['my-operator.example.com', 'x' * 63, 'my-operator.example.com/' + 'x' * 63],
    ['my-operator.example.com', 'x' * 64, 'my-operator.example.com/' + 'x' * 54 + 'xx-SItAqA'],
    ['my-operator.example.com', 'y' * 64, 'my-operator.example.com/' + 'y' * 54 + 'yy-0d251g'],
    ['my-operator.example.com', 'z' * 64, 'my-operator.example.com/' + 'z' * 54 + 'zz-E7wvIA'],

    # For special chars in base64 encoding ("+" and "/"), which are not compatible with K8s.
    # The numbers are found empirically so that both "/" and "+" are found in the base64'ed digest.
    ['my-operator.example.com', 'fn' * 323, 'my-operator.example.com/' + 'fn' * 27 + 'fn-Az-r.g'],
]


@pytest.mark.parametrize('cls', STORAGE_KEY_FORMING_CLASSES)
def test_keys_for_all_versions(cls):
    storage = cls(v1=True, prefix='kopf.zalando.org')
    v1_key = storage.make_v1_key('.' * 64)
    v2_key = storage.make_v2_key('.' * 64)
    assert v1_key != v2_key  # prerequisite
    keys = storage.make_keys('.' * 64)
    assert len(list(keys)) == 2
    assert v1_key in keys
    assert v2_key in keys


@pytest.mark.parametrize('cls', STORAGE_KEY_FORMING_CLASSES)
def test_keys_deduplication(cls):
    storage = cls(v1=True, prefix='kopf.zalando.org')
    v1_key = storage.make_v1_key('...')
    v2_key = storage.make_v2_key('...')
    assert v1_key == v2_key  # prerequisite
    keys = storage.make_keys('...')
    assert len(list(keys)) == 1
    assert v1_key in keys
    assert v2_key in keys


@pytest.mark.parametrize('kind, owners, expected', [
    pytest.param('ReplicaSet', [{'kind': 'Deployment'}], 'kopf.dev/xyz-ofDRS', id='DRS'),
    pytest.param('ReplicaSet', [{'kind': 'OtherOwner'}], 'kopf.dev/xyz', id='not-deployment'),
    pytest.param('OtherKind', [{'kind': 'Deployment'}], 'kopf.dev/xyz', id='not-replicaset'),
])
@pytest.mark.parametrize('cls', STORAGE_KEY_FORMING_CLASSES)
def test_keys_of_replicaset_owned_by_deployment(cls, kind, owners, expected):
    storage = cls(v1=True, prefix='kopf.dev')
    body = Body({'kind': kind, 'metadata': {'ownerReferences': owners}})
    keys = storage.make_keys('xyz', body=body)
    assert set(keys) == {expected}


@pytest.mark.parametrize('prefix, provided_key, expected_key', COMMON_KEYS + V1_KEYS)
@pytest.mark.parametrize('cls', STORAGE_KEY_FORMING_CLASSES)
def test_key_hashing_v1(cls, prefix, provided_key, expected_key):
    storage = cls(v1=True, prefix=prefix)
    returned_key = storage.make_v1_key(provided_key)
    assert returned_key == expected_key


@pytest.mark.parametrize('prefix, provided_key, expected_key', COMMON_KEYS + V2_KEYS)
@pytest.mark.parametrize('cls', STORAGE_KEY_FORMING_CLASSES)
def test_key_hashing_v2(cls, prefix, provided_key, expected_key):
    storage = cls(v1=True, prefix=prefix)
    returned_key = storage.make_v2_key(provided_key)
    assert returned_key == expected_key


@pytest.mark.parametrize('prefix, provided_key, expected_key', COMMON_KEYS + V1_KEYS + V2_KEYS)
@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_keys_hashed_on_fetching(cls, prefix, provided_key, expected_key):
    storage = cls(prefix=prefix)
    body = Body({'metadata': {'annotations': {expected_key: CONTENT_JSON}}})
    record = storage.fetch(body=body, key=HandlerId(provided_key))
    assert record is not None
    assert record == CONTENT_DATA


@pytest.mark.parametrize('prefix, provided_key, expected_key', COMMON_KEYS + V1_KEYS + V2_KEYS)
@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_keys_normalized_on_storing(cls, prefix, provided_key, expected_key):
    storage = cls(prefix=prefix)
    patch = Patch()
    body = Body({'metadata': {'annotations': {expected_key: 'null'}}})
    storage.store(body=body, patch=patch, key=HandlerId(provided_key), record=CONTENT_DATA)
    assert expected_key in patch.metadata.annotations


@pytest.mark.parametrize('prefix, provided_key, expected_key', COMMON_KEYS + V1_KEYS + V2_KEYS)
@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_keys_normalized_on_purging(cls, prefix, provided_key, expected_key):
    storage = cls(prefix=prefix)
    patch = Patch()
    body = Body({'metadata': {'annotations': {expected_key: 'null'}}})
    storage.purge(body=body, patch=patch, key=HandlerId(provided_key))
    assert expected_key in patch.metadata.annotations


@pytest.mark.parametrize('prefix, provided_key, expected_key', COMMON_KEYS + V1_KEYS + V2_KEYS)
@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES)
def test_keys_normalized_on_touching(cls, prefix, provided_key, expected_key):
    storage = cls(prefix=prefix, touch_key=provided_key)
    patch = Patch()
    body = Body({})
    storage.touch(body=body, patch=patch, value='irrelevant')
    assert expected_key in patch.metadata.annotations


@pytest.mark.parametrize('cls', ANNOTATIONS_POPULATING_STORAGES + STORAGE_KEY_FORMING_CLASSES)
def test_warning_on_long_prefix(cls):
    with pytest.warns(UserWarning, match=r"The annotations prefix is too long"):
        cls(v1=True, prefix='x' * (253 - 63))



import dataclasses
import enum
from typing import Any, NewType, Optional

from kopf.structs import callbacks, dicts, filters, references

# Strings are taken from the users, but then tainted as this type for stricter type-checking:
# to prevent usage of some other strings (e.g. operator id) as the handlers ids.
HandlerId = NewType('HandlerId', str)


class ErrorsMode(enum.Enum):
    """ How arbitrary (non-temporary/non-permanent) exceptions are treated. """
    IGNORED = enum.auto()
    TEMPORARY = enum.auto()
    PERMANENT = enum.auto()


class Activity(str, enum.Enum):
    STARTUP = 'startup'
    CLEANUP = 'cleanup'
    AUTHENTICATION = 'authentication'
    PROBE = 'probe'


# Constants for cause types, to prevent a direct usage of strings, and typos.
# They are not exposed by the framework, but are used internally. See also: `kopf.on`.
class Reason(str, enum.Enum):
    CREATE = 'create'
    UPDATE = 'update'
    DELETE = 'delete'
    RESUME = 'resume'
    NOOP = 'noop'
    FREE = 'free'
    GONE = 'gone'

    def __str__(self) -> str:
        return str(self.value)


# These sets are checked in few places, so we keep them centralised:
# the user-facing causes (for handlers) and internally facing (for the reactor).
HANDLER_REASONS = (
    Reason.CREATE,
    Reason.UPDATE,
    Reason.DELETE,
    Reason.RESUME,
)
REACTOR_REASONS = (
    Reason.NOOP,
    Reason.FREE,
    Reason.GONE,
)
ALL_REASONS = HANDLER_REASONS + REACTOR_REASONS

# The human-readable names of these causes. Will be capitalised when needed.
TITLES = {
    Reason.CREATE: 'creation',
    Reason.UPDATE: 'updating',
    Reason.DELETE: 'deletion',
    Reason.RESUME: 'resuming',
}


# A registered handler (function + meta info).
# FIXME: Must be frozen, but mypy fails in _call_handler() with a cryptic error:
# FIXME:    Argument 1 to "invoke" has incompatible type "Optional[HandlerResult]";
# FIXME:    expected "Union[LifeCycleFn, ActivityHandlerFn, ResourceHandlerFn]"
@dataclasses.dataclass
class BaseHandler:
    id: HandlerId
    fn: callbacks.BaseFn
    param: Optional[Any]
    errors: Optional[ErrorsMode]
    timeout: Optional[float]
    retries: Optional[int]
    backoff: Optional[float]

    # Used in the logs. Overridden in some (but not all) handler types for better log messages.
    def __str__(self) -> str:
        return f"Handler {self.id!r}"


@dataclasses.dataclass
class ActivityHandler(BaseHandler):
    fn: callbacks.ActivityFn  # type clarification
    activity: Optional[Activity]
    _fallback: bool = False  # non-public!

    def __str__(self) -> str:
        return f"Activity {self.id!r}"


@dataclasses.dataclass
class ResourceHandler(BaseHandler):
    selector: Optional[references.Selector]  # None is used only in sub-handlers
    labels: Optional[filters.MetaFilter]
    annotations: Optional[filters.MetaFilter]
    when: Optional[callbacks.WhenFilterFn]
    field: Optional[dicts.FieldPath]
    value: Optional[filters.ValueFilter]

    @property
    def requires_patching(self) -> bool:
        return True  # all typical handlers except several ones with overrides


@dataclasses.dataclass
class ResourceIndexingHandler(ResourceHandler):
    fn: callbacks.ResourceIndexingFn  # type clarification

    def __str__(self) -> str:
        return f"Indexer {self.id!r}"

    @property
    def requires_patching(self) -> bool:
        return False


@dataclasses.dataclass
class ResourceWatchingHandler(ResourceHandler):
    fn: callbacks.ResourceWatchingFn  # type clarification

    @property
    def requires_patching(self) -> bool:
        return False


@dataclasses.dataclass
class ResourceChangingHandler(ResourceHandler):
    fn: callbacks.ResourceChangingFn  # type clarification
    reason: Optional[Reason]
    initial: Optional[bool]
    deleted: Optional[bool]  # used for mixed-in (initial==True) @on.resume handlers only.
    requires_finalizer: Optional[bool]
    field_needs_change: Optional[bool]  # to identify on-field/on-update with support for old=/new=.
    old: Optional[filters.ValueFilter]
    new: Optional[filters.ValueFilter]


@dataclasses.dataclass
class ResourceSpawningHandler(ResourceHandler):
    requires_finalizer: Optional[bool]
    initial_delay: Optional[float]


@dataclasses.dataclass
class ResourceDaemonHandler(ResourceSpawningHandler):
    fn: callbacks.ResourceDaemonFn  # type clarification
    cancellation_backoff: Optional[float]  # how long to wait before actual cancellation.
    cancellation_timeout: Optional[float]  # how long to wait before giving up on cancellation.
    cancellation_polling: Optional[float]  # how often to check for cancellation status.

    def __str__(self) -> str:
        return f"Daemon {self.id!r}"


@dataclasses.dataclass
class ResourceTimerHandler(ResourceSpawningHandler):
    fn: callbacks.ResourceTimerFn  # type clarification
    sharp: Optional[bool]
    idle: Optional[float]
    interval: Optional[float]

    def __str__(self) -> str:
        return f"Timer {self.id!r}"



import pytest

import kopf
from kopf.structs.bodies import Body


def test_in_owner_reference_appending():
    with pytest.raises(TypeError) as e:
        kopf.append_owner_reference(object(), Body({}))
    assert "K8s object class is not supported" in str(e.value)


def test_in_owner_reference_removal():
    with pytest.raises(TypeError) as e:
        kopf.remove_owner_reference(object(), Body({}))
    assert "K8s object class is not supported" in str(e.value)


def test_in_name_harmonization():
    with pytest.raises(TypeError) as e:
        kopf.harmonize_naming(object(), 'x')
    assert "K8s object class is not supported" in str(e.value)


def test_in_namepace_adjustment():
    with pytest.raises(TypeError) as e:
        kopf.adjust_namespace(object(), 'x')
    assert "K8s object class is not supported" in str(e.value)


def test_in_labelling():
    with pytest.raises(TypeError) as e:
        kopf.label(object(), {})
    assert "K8s object class is not supported" in str(e.value)


def test_in_adopting():
    with pytest.raises(TypeError) as e:
        kopf.adopt(object(), Body({}))
    assert "K8s object class is not supported" in str(e.value)



import kopf

E2E_TRACEBACKS = True
E2E_CREATION_STOP_WORDS = ['Something has changed,']
E2E_SUCCESS_COUNTS = {'eventual_success_with_few_messages': 1}
E2E_FAILURE_COUNTS = {'eventual_failure_with_tracebacks': 1, 'instant_failure_with_traceback': 1, 'instant_failure_with_only_a_message': 1}


class MyException(Exception):
    pass


@kopf.on.create('kopfexamples')
def instant_failure_with_only_a_message(**kwargs):
    raise kopf.PermanentError("Fail once and for all.")


@kopf.on.create('kopfexamples')
def eventual_success_with_few_messages(retry, **kwargs):
    if retry < 3:  # 0, 1, 2, 3
        raise kopf.TemporaryError("Expected recoverable error.", delay=1.0)


@kopf.on.create('kopfexamples', retries=3, backoff=1.0)
def eventual_failure_with_tracebacks(**kwargs):
    raise MyException("An error that is supposed to be recoverable.")


@kopf.on.create('kopfexamples', errors=kopf.ErrorsMode.PERMANENT, backoff=1.0)
def instant_failure_with_traceback(**kwargs):
    raise MyException("An error that is supposed to be recoverable.")



"""
Only the tests from the K8s client library (simulated) to the event stream.

Excluded: the watch-stream consumption and queue multiplexing routines
(see ``tests_queueing.py``).

Used for internal control that the event multiplexing works are intended.
If the intentions change, the tests should be rewritten.
They are NOT part of the public interface of the framework.
"""
import asyncio
import logging

import pytest

from kopf.clients.watching import Bookmark, WatchingError, continuous_watch

STREAM_WITH_NORMAL_EVENTS = [
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
]
STREAM_WITH_UNKNOWN_EVENT = [
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'UNKNOWN', 'object': {}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
]
STREAM_WITH_ERROR_410GONE = [
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'ERROR', 'object': {'code': 410}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
]
STREAM_WITH_ERROR_CODE = [
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'ERROR', 'object': {'code': 666}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
]


class SampleException(Exception):
    pass


async def test_empty_stream_yields_nothing(
        settings, resource, stream, namespace):

    stream.feed([], namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    async for event in continuous_watch(settings=settings,
                                        resource=resource,
                                        namespace=namespace,
                                        operator_pause_waiter=asyncio.Future()):
        events.append(event)

    assert len(events) == 1
    assert events[0] == Bookmark.LISTED


async def test_event_stream_yields_everything(
        settings, resource, stream, namespace):

    stream.feed(STREAM_WITH_NORMAL_EVENTS, namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    async for event in continuous_watch(settings=settings,
                                        resource=resource,
                                        namespace=namespace,
                                        operator_pause_waiter=asyncio.Future()):
        events.append(event)

    assert len(events) == 3
    assert events[0] == Bookmark.LISTED
    assert events[1]['object']['spec'] == 'a'
    assert events[2]['object']['spec'] == 'b'


async def test_unknown_event_type_ignored(
        settings, resource, stream, namespace, caplog):
    caplog.set_level(logging.DEBUG)

    stream.feed(STREAM_WITH_UNKNOWN_EVENT, namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    async for event in continuous_watch(settings=settings,
                                        resource=resource,
                                        namespace=namespace,
                                        operator_pause_waiter=asyncio.Future()):
        events.append(event)

    assert len(events) == 3
    assert events[0] == Bookmark.LISTED
    assert events[1]['object']['spec'] == 'a'
    assert events[2]['object']['spec'] == 'b'
    assert "Ignoring an unsupported event type" in caplog.text
    assert "UNKNOWN" in caplog.text


async def test_error_410gone_exits_normally(
        settings, resource, stream, namespace, caplog):
    caplog.set_level(logging.DEBUG)

    stream.feed(STREAM_WITH_ERROR_410GONE, namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    async for event in continuous_watch(settings=settings,
                                        resource=resource,
                                        namespace=namespace,
                                        operator_pause_waiter=asyncio.Future()):
        events.append(event)

    assert len(events) == 2
    assert events[0] == Bookmark.LISTED
    assert events[1]['object']['spec'] == 'a'
    assert "Restarting the watch-stream" in caplog.text


async def test_unknown_error_raises_exception(
        settings, resource, stream, namespace):

    stream.feed(STREAM_WITH_ERROR_CODE, namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    with pytest.raises(WatchingError) as e:
        async for event in continuous_watch(settings=settings,
                                            resource=resource,
                                            namespace=namespace,
                                            operator_pause_waiter=asyncio.Future()):
            events.append(event)

    assert len(events) == 2
    assert events[0] == Bookmark.LISTED
    assert events[1]['object']['spec'] == 'a'
    assert '666' in str(e.value)


async def test_exception_escalates(
        settings, resource, stream, namespace, enforced_session, mocker):

    enforced_session.get = mocker.Mock(side_effect=SampleException())
    stream.feed([], namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    with pytest.raises(SampleException):
        async for event in continuous_watch(settings=settings,
                                            resource=resource,
                                            namespace=namespace,
                                            operator_pause_waiter=asyncio.Future()):
            events.append(event)

    assert len(events) == 0


# See: See: https://github.com/zalando-incubator/kopf/issues/275
async def test_long_line_parsing(
        settings, resource, stream, namespace, aresponses):

    content = [
        {'type': 'ADDED', 'object': {'spec': {'field': 'x'}}},
        {'type': 'ADDED', 'object': {'spec': {'field': 'y' * (2 * 1024 * 1024)}}},
        {'type': 'ADDED', 'object': {'spec': {'field': 'z' * (4 * 1024 * 1024)}}},
    ]
    stream.feed(content, namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    async for event in continuous_watch(settings=settings,
                                        resource=resource,
                                        namespace=namespace,
                                        operator_pause_waiter=asyncio.Future()):
        events.append(event)

    assert len(events) == 4
    assert events[0] == Bookmark.LISTED
    assert len(events[1]['object']['spec']['field']) == 1
    assert len(events[2]['object']['spec']['field']) == 2 * 1024 * 1024
    assert len(events[3]['object']['spec']['field']) == 4 * 1024 * 1024



import glob
import os.path
import pathlib
import subprocess

import pytest

from kopf.reactor.registries import SmartOperatorRegistry

root_dir = os.path.relpath(os.path.dirname(os.path.dirname(os.path.dirname(__file__))))
examples = sorted(glob.glob(os.path.join(root_dir, 'examples/*/')))
assert examples  # if empty, it is just the detection failed
examples = [path for path in examples if not glob.glob((os.path.join(path, 'test*.py')))]


@pytest.fixture
def registry_factory():
    # Authentication is needed for the real e2e tests.
    return SmartOperatorRegistry


@pytest.fixture(params=examples, ids=[os.path.basename(path.rstrip('/')) for path in examples])
def exampledir(request):
    return pathlib.Path(request.param)


@pytest.fixture(scope='session')
def peering_yaml():
    crd_api = os.environ.get('CRDAPI') or 'v1'
    crd_file = 'peering.yaml' if crd_api == 'v1' else f'peering-{crd_api}.yaml'
    return f'{crd_file}'


@pytest.fixture(scope='session')
def crd_yaml():
    crd_api = os.environ.get('CRDAPI') or 'v1'
    crd_file = 'crd.yaml' if crd_api == 'v1' else f'crd-{crd_api}.yaml'
    return f'examples/{crd_file}'


@pytest.fixture()
def with_crd(crd_yaml):
    # Our best guess on which Kubernetes version we are running on.
    subprocess.run(f"kubectl apply -f {crd_yaml}",
                   shell=True, check=True, timeout=10, capture_output=True)


@pytest.fixture()
def with_peering(peering_yaml):
    subprocess.run(f"kubectl apply -f {peering_yaml}",
                   shell=True, check=True, timeout=10, capture_output=True)


@pytest.fixture()
def no_crd():
    subprocess.run("kubectl delete customresourcedefinition kopfexamples.kopf.dev",
                   shell=True, check=True, timeout=10, capture_output=True)


@pytest.fixture()
def no_peering():
    subprocess.run("kubectl delete customresourcedefinition kopfpeerings.kopf.dev",
                   shell=True, check=True, timeout=10, capture_output=True)



import copy

import pytest

import kopf
from kopf.reactor.causation import ResourceWatchingCause
from kopf.structs.dicts import parse_field
from kopf.structs.filters import ABSENT, PRESENT
from kopf.structs.handlers import ResourceWatchingHandler


# Used in the tests. Must be global-scoped, or its qualname will be affected.
def some_fn(x=None):
    pass


def _never(*_, **__):
    return False


def _always(*_, **__):
    return True


@pytest.fixture()
def handler_factory(registry, selector):
    def factory(**kwargs):
        handler = ResourceWatchingHandler(**dict(dict(
            fn=some_fn, id='a', param=None,
            errors=None, timeout=None, retries=None, backoff=None,
            selector=selector, annotations=None, labels=None, when=None,
            field=None, value=None,
        ), **kwargs))
        registry._resource_watching.append(handler)
        return handler
    return factory


@pytest.fixture(params=[
    pytest.param(dict(body={}), id='no-field'),
])
def cause_no_field(request, cause_factory):
    kwargs = copy.deepcopy(request.param)
    kwargs['body'].update({'metadata': {'labels': {'known': 'value'},
                                        'annotations': {'known': 'value'}}})
    cause = cause_factory(cls=ResourceWatchingCause, **kwargs)
    return cause


@pytest.fixture(params=[
    pytest.param(dict(body={'known-field': 'new'}), id='with-field'),
])
def cause_with_field(request, cause_factory):
    kwargs = copy.deepcopy(request.param)
    kwargs['body'].update({'metadata': {'labels': {'known': 'value'},
                                        'annotations': {'known': 'value'}}})
    cause = cause_factory(cls=ResourceWatchingCause, **kwargs)
    return cause


@pytest.fixture(params=[
    # The original no-diff was equivalent to no-field until body/old/new were added to the check.
    pytest.param(dict(body={}, diff=[]), id='no-field'),
    pytest.param(dict(body={'known-field': 'new'}), id='with-field'),
])
def cause_any_field(request, cause_factory):
    kwargs = copy.deepcopy(request.param)
    kwargs['body'].update({'metadata': {'labels': {'known': 'value'},
                                        'annotations': {'known': 'value'}}})
    cause = cause_factory(cls=ResourceWatchingCause, **kwargs)
    return cause


#
# "Catch-all" handlers are those with event == None.
#

def test_catchall_handlers_without_field_found(
        cause_any_field, registry, handler_factory):
    cause = cause_any_field
    handler_factory(field=None)
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


def test_catchall_handlers_with_field_found(
        cause_with_field, registry, handler_factory):
    cause = cause_with_field
    handler_factory(field=parse_field('known-field'))
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


def test_catchall_handlers_with_field_ignored(
        cause_no_field, registry, handler_factory):
    cause = cause_no_field
    handler_factory(field=parse_field('known-field'))
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers


@pytest.mark.parametrize('labels', [
    pytest.param({'known': 'value'}, id='with-label'),
    pytest.param({'known': 'value', 'extra': 'other'}, id='with-extra-label'),
])
def test_catchall_handlers_with_exact_labels_satisfied(
        cause_factory, registry, handler_factory, labels):
    cause = cause_factory(body={'metadata': {'labels': labels}})
    handler_factory(labels={'known': 'value'})
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


@pytest.mark.parametrize('labels', [
    pytest.param({}, id='without-label'),
    pytest.param({'known': 'other'}, id='with-other-value'),
    pytest.param({'extra': 'other'}, id='with-other-label'),
])
def test_catchall_handlers_with_exact_labels_not_satisfied(
        cause_factory, registry, handler_factory, labels):
    cause = cause_factory(body={'metadata': {'labels': labels}})
    handler_factory(labels={'known': 'value'})
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers


@pytest.mark.parametrize('labels', [
    pytest.param({'known': 'value'}, id='with-label'),
    pytest.param({'known': 'other'}, id='with-other-value'),
])
def test_catchall_handlers_with_desired_labels_present(
        cause_factory, registry, handler_factory, labels):
    cause = cause_factory(body={'metadata': {'labels': labels}})
    handler_factory(labels={'known': PRESENT})
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


@pytest.mark.parametrize('labels', [
    pytest.param({}, id='without-label'),
    pytest.param({'extra': 'other'}, id='with-other-label'),
])
def test_catchall_handlers_with_desired_labels_absent(
        cause_factory, registry, handler_factory, labels):
    cause = cause_factory(body={'metadata': {'labels': labels}})
    handler_factory(labels={'known': PRESENT})
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers


@pytest.mark.parametrize('labels', [
    pytest.param({'known': 'value'}, id='with-label'),
    pytest.param({'known': 'other'}, id='with-other-value'),
])
def test_catchall_handlers_with_undesired_labels_present(
        cause_factory, registry, handler_factory, labels):
    cause = cause_factory(body={'metadata': {'labels': labels}})
    handler_factory(labels={'known': ABSENT})
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers


@pytest.mark.parametrize('labels', [
    pytest.param({}, id='without-label'),
    pytest.param({'extra': 'other'}, id='with-other-label'),
])
def test_catchall_handlers_with_undesired_labels_absent(
        cause_factory, registry, handler_factory, labels):
    cause = cause_factory(body={'metadata': {'labels': labels}})
    handler_factory(labels={'known': ABSENT})
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


@pytest.mark.parametrize('labels', [
    pytest.param({}, id='without-label'),
    pytest.param({'known': 'value'}, id='with-label'),
    pytest.param({'known': 'other'}, id='with-other-value'),
])
def test_catchall_handlers_with_labels_callback_says_true(
        cause_factory, registry, handler_factory, labels):
    cause = cause_factory(body={'metadata': {'labels': labels}})
    handler_factory(labels={'known': _always})
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


@pytest.mark.parametrize('labels', [
    pytest.param({}, id='without-label'),
    pytest.param({'known': 'value'}, id='with-label'),
    pytest.param({'known': 'other'}, id='with-other-value'),
])
def test_catchall_handlers_with_labels_callback_says_false(
        cause_factory, registry, handler_factory, labels):
    cause = cause_factory(body={'metadata': {'labels': labels}})
    handler_factory(labels={'known': _never})
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers


@pytest.mark.parametrize('labels', [
    pytest.param({}, id='without-label'),
    pytest.param({'known': 'value'}, id='with-label'),
    pytest.param({'known': 'other'}, id='with-other-value'),
    pytest.param({'extra': 'other'}, id='with-other-label'),
    pytest.param({'known': 'value', 'extra': 'other'}, id='with-extra-label'),
])
def test_catchall_handlers_without_labels(
        cause_factory, registry, handler_factory, labels):
    cause = cause_factory(body={'metadata': {'labels': labels}})
    handler_factory(labels=None)
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


@pytest.mark.parametrize('annotations', [
    pytest.param({'known': 'value'}, id='with-annotation'),
    pytest.param({'known': 'value', 'extra': 'other'}, id='with-extra-annotation'),
])
def test_catchall_handlers_with_exact_annotations_satisfied(
        cause_factory, registry, handler_factory, annotations):
    cause = cause_factory(body={'metadata': {'annotations': annotations}})
    handler_factory(annotations={'known': 'value'})
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


@pytest.mark.parametrize('annotations', [
    pytest.param({}, id='without-annotation'),
    pytest.param({'known': 'other'}, id='with-other-value'),
    pytest.param({'extra': 'other'}, id='with-other-annotation'),
])
def test_catchall_handlers_with_exact_annotations_not_satisfied(
        cause_factory, registry, handler_factory, annotations):
    cause = cause_factory(body={'metadata': {'annotations': annotations}})
    handler_factory(annotations={'known': 'value'})
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers


@pytest.mark.parametrize('annotations', [
    pytest.param({'known': 'value'}, id='with-annotation'),
    pytest.param({'known': 'other'}, id='with-other-value'),
])
def test_catchall_handlers_with_desired_annotations_present(
        cause_factory, registry, handler_factory, annotations):
    cause = cause_factory(body={'metadata': {'annotations': annotations}})
    handler_factory(annotations={'known': PRESENT})
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


@pytest.mark.parametrize('annotations', [
    pytest.param({}, id='without-annotation'),
    pytest.param({'extra': 'other'}, id='with-other-annotation'),
])
def test_catchall_handlers_with_desired_annotations_absent(
        cause_factory, registry, handler_factory, annotations):
    cause = cause_factory(body={'metadata': {'annotations': annotations}})
    handler_factory(annotations={'known': PRESENT})
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers


@pytest.mark.parametrize('annotations', [
    pytest.param({'known': 'value'}, id='with-annotation'),
    pytest.param({'known': 'other'}, id='with-other-value'),
])
def test_catchall_handlers_with_undesired_annotations_present(
        cause_factory, registry, handler_factory, annotations):
    cause = cause_factory(body={'metadata': {'annotations': annotations}})
    handler_factory(annotations={'known': ABSENT})
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers


@pytest.mark.parametrize('annotations', [
    pytest.param({}, id='without-annotation'),
    pytest.param({'extra': 'other'}, id='with-other-annotation'),
])
def test_catchall_handlers_with_undesired_annotations_absent(
        cause_factory, registry, handler_factory, annotations):
    cause = cause_factory(body={'metadata': {'annotations': annotations}})
    handler_factory(annotations={'known': ABSENT})
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


@pytest.mark.parametrize('annotations', [
    pytest.param({}, id='without-annotation'),
    pytest.param({'known': 'value'}, id='with-annotation'),
    pytest.param({'known': 'other'}, id='with-other-value'),
])
def test_catchall_handlers_with_annotations_callback_says_true(
        cause_factory, registry, handler_factory, annotations):
    cause = cause_factory(body={'metadata': {'annotations': annotations}})
    handler_factory(annotations={'known': _always})
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


@pytest.mark.parametrize('annotations', [
    pytest.param({}, id='without-annotation'),
    pytest.param({'known': 'value'}, id='with-annotation'),
    pytest.param({'known': 'other'}, id='with-other-value'),
])
def test_catchall_handlers_with_annotations_callback_says_false(
        cause_factory, registry, handler_factory, annotations):
    cause = cause_factory(body={'metadata': {'annotations': annotations}})
    handler_factory(annotations={'known': _never})
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers


@pytest.mark.parametrize('annotations', [
    pytest.param({}, id='without-annotation'),
    pytest.param({'known': 'value'}, id='with-annotation'),
    pytest.param({'known': 'other'}, id='with-other-value'),
    pytest.param({'extra': 'other'}, id='with-other-annotation'),
    pytest.param({'known': 'value', 'extra': 'other'}, id='with-extra-annotation'),
])
def test_catchall_handlers_without_annotations(
        cause_factory, registry, handler_factory, annotations):
    cause = cause_factory(body={'metadata': {'annotations': annotations}})
    handler_factory()
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


@pytest.mark.parametrize('labels, annotations', [
    pytest.param({'known': 'value'}, {'known': 'value'}, id='with-label-annotation'),
    pytest.param({'known': 'value', 'extra': 'other'}, {'known': 'value'}, id='with-extra-label-annotation'),
    pytest.param({'known': 'value'}, {'known': 'value', 'extra': 'other'}, id='with-label-extra-annotation'),
    pytest.param({'known': 'value', 'extra': 'other'}, {'known': 'value', 'extra': 'other'}, id='with-extra-label-extra-annotation'),
])
def test_catchall_handlers_with_labels_and_annotations_satisfied(
        cause_factory, registry, handler_factory, labels, annotations):
    cause = cause_factory(body={'metadata': {'labels': labels, 'annotations': annotations}})
    handler_factory(labels={'known': 'value'}, annotations={'known': 'value'})
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


@pytest.mark.parametrize('labels', [
    pytest.param({}, id='without-label'),
    pytest.param({'known': 'value'}, id='with-label'),
    pytest.param({'known': 'other'}, id='with-other-value'),
    pytest.param({'extra': 'other'}, id='with-other-label'),
    pytest.param({'known': 'value', 'extra': 'other'}, id='with-extra-label'),
])
def test_catchall_handlers_with_labels_and_annotations_not_satisfied(
        cause_factory, registry, handler_factory, labels):
    cause = cause_factory(body={'metadata': {'labels': labels}})
    handler_factory(labels={'known': 'value'}, annotations={'known': 'value'})
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers


@pytest.mark.parametrize('when', [
    pytest.param(None, id='without-when'),
    pytest.param(lambda body=None, **_: body['spec']['name'] == 'test', id='with-when'),
    pytest.param(lambda **_: True, id='with-other-when'),
])
def test_catchall_handlers_with_when_callback_matching(
        cause_factory, registry, handler_factory, when):
    cause = cause_factory(body={'spec': {'name': 'test'}})
    handler_factory(when=when)
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


@pytest.mark.parametrize('when', [
    pytest.param(lambda body=None, **_: body['spec']['name'] != "test", id='with-when'),
    pytest.param(lambda **_: False, id='with-other-when'),
])
def test_catchall_handlers_with_when_callback_mismatching(
        cause_factory, registry, handler_factory, when):
    cause = cause_factory(body={'spec': {'name': 'test'}})
    handler_factory(when=when)
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers


def test_decorator_without_field_found(
        cause_any_field, registry, resource):

    @kopf.on.event(*resource, field=None)
    def some_fn(**_): ...

    cause = cause_any_field
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


def test_decorator_with_field_found(
        cause_with_field, registry, resource):

    @kopf.on.event(*resource, field='known-field')
    def some_fn(**_): ...

    cause = cause_with_field
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


def test_decorator_with_field_ignored(
        cause_no_field, registry, resource):

    @kopf.on.event(*resource, field='known-field')
    def some_fn(**_): ...

    cause = cause_no_field
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers


def test_decorator_with_labels_satisfied(
        cause_any_field, registry, resource):

    @kopf.on.event(*resource, labels={'known': PRESENT})
    def some_fn(**_): ...

    cause = cause_any_field
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


def test_decorator_with_labels_not_satisfied(
        cause_any_field, registry, resource):

    @kopf.on.event(*resource, labels={'extra': PRESENT})
    def some_fn(**_): ...

    cause = cause_any_field
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers


def test_decorator_with_annotations_satisfied(
        cause_any_field, registry, resource):

    @kopf.on.event(*resource, annotations={'known': PRESENT})
    def some_fn(**_): ...

    cause = cause_any_field
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


def test_decorator_with_annotations_not_satisfied(
        cause_any_field, registry, resource):

    @kopf.on.event(*resource, annotations={'extra': PRESENT})
    def some_fn(**_): ...

    cause = cause_any_field
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers


def test_decorator_with_filter_satisfied(
        cause_any_field, registry, resource):

    @kopf.on.event(*resource, when=_always)
    def some_fn(**_): ...

    cause = cause_any_field
    handlers = registry._resource_watching.get_handlers(cause)
    assert handlers


def test_decorator_with_filter_not_satisfied(
        cause_any_field, registry, resource):

    @kopf.on.event(*resource, when=_never)
    def some_fn(**_): ...

    cause = cause_any_field
    handlers = registry._resource_watching.get_handlers(cause)
    assert not handlers



import collections
import re
import subprocess
import time
from typing import Any, Optional, Sequence

import pytest

from kopf.testing import KopfRunner


def test_all_examples_are_runnable(mocker, settings, with_crd, exampledir, caplog):

    # If the example has its own opinion on the timing, try to respect it.
    # See e.g. /examples/99-all-at-once/example.py.
    example_py = exampledir / 'example.py'
    e2e_startup_time_limit = _parse_e2e_value(str(example_py), 'E2E_STARTUP_TIME_LIMIT')
    e2e_startup_stop_words = _parse_e2e_value(str(example_py), 'E2E_STARTUP_STOP_WORDS')
    e2e_cleanup_time_limit = _parse_e2e_value(str(example_py), 'E2E_CLEANUP_TIME_LIMIT')
    e2e_cleanup_stop_words = _parse_e2e_value(str(example_py), 'E2E_CLEANUP_STOP_WORDS')
    e2e_creation_time_limit = _parse_e2e_value(str(example_py), 'E2E_CREATION_TIME_LIMIT')
    e2e_creation_stop_words = _parse_e2e_value(str(example_py), 'E2E_CREATION_STOP_WORDS')
    e2e_deletion_time_limit = _parse_e2e_value(str(example_py), 'E2E_DELETION_TIME_LIMIT')
    e2e_deletion_stop_words = _parse_e2e_value(str(example_py), 'E2E_DELETION_STOP_WORDS')
    e2e_tracebacks = _parse_e2e_value(str(example_py), 'E2E_TRACEBACKS')
    e2e_success_counts = _parse_e2e_value(str(example_py), 'E2E_SUCCESS_COUNTS')
    e2e_failure_counts = _parse_e2e_value(str(example_py), 'E2E_FAILURE_COUNTS')
    e2e_test_creation = _parse_e2e_presence(str(example_py), r'@kopf.on.create\(')
    e2e_test_highlevel = _parse_e2e_presence(str(example_py), r'@kopf.on.(create|update|delete)\(')

    # check whether there are mandatory deletion handlers or not
    m = re.search(r'@kopf\.on\.delete\((\s|.*)?(optional=(\w+))?\)', example_py.read_text(), re.M)
    requires_finalizer = False
    if m:
        requires_finalizer = True
        if m.group(2):
            requires_finalizer = not eval(m.group(3))

    # To prevent lengthy sleeps on the simulated retries.
    mocker.patch('kopf.reactor.handling.DEFAULT_RETRY_DELAY', 1)

    # To prevent lengthy threads in the loop executor when the process exits.
    settings.watching.server_timeout = 10

    # Run an operator and simulate some activity with the operated resource.
    with KopfRunner(
        ['run', '--all-namespaces', '--standalone', '--verbose', str(example_py)],
        timeout=60,
    ) as runner:

        # Give it some time to start.
        _sleep_till_stopword(caplog=caplog,
                             delay=e2e_startup_time_limit,
                             patterns=e2e_startup_stop_words or ['Client is configured'])

        # Trigger the reaction. Give it some time to react and to sleep and to retry.
        subprocess.run("kubectl apply -f examples/obj.yaml",
                       shell=True, check=True, timeout=10, capture_output=True)
        _sleep_till_stopword(caplog=caplog,
                             delay=e2e_creation_time_limit,
                             patterns=e2e_creation_stop_words)

        # Trigger the reaction. Give it some time to react.
        subprocess.run("kubectl delete -f examples/obj.yaml",
                       shell=True, check=True, timeout=10, capture_output=True)
        _sleep_till_stopword(caplog=caplog,
                             delay=e2e_deletion_time_limit,
                             patterns=e2e_deletion_stop_words)

    # Give it some time to finish.
    _sleep_till_stopword(caplog=caplog,
                         delay=e2e_cleanup_time_limit,
                         patterns=e2e_cleanup_stop_words or ['Hung tasks', 'Root tasks'])

    # Verify that the operator did not die on start, or during the operation.
    assert runner.exception is None
    assert runner.exit_code == 0

    # There are usually more than these messages, but we only check for the certain ones.
    # This just shows us that the operator is doing something, it is alive.
    if requires_finalizer:
        assert '[default/kopf-example-1] Adding the finalizer' in runner.stdout
    if e2e_test_creation:
        assert '[default/kopf-example-1] Creation is in progress:' in runner.stdout
    if requires_finalizer:
        assert '[default/kopf-example-1] Deletion is in progress:' in runner.stdout
    if e2e_test_highlevel:
        assert '[default/kopf-example-1] Deleted, really deleted' in runner.stdout
    if not e2e_tracebacks:
        assert 'Traceback (most recent call last):' not in runner.stdout

    # Verify that once a handler succeeds, it is never re-executed again.
    handler_names = re.findall(r"'(.+?)' succeeded", runner.stdout)
    if e2e_success_counts is not None:
        checked_names = [name for name in handler_names if name in e2e_success_counts]
        name_counts = collections.Counter(checked_names)
        assert name_counts == e2e_success_counts
    else:
        name_counts = collections.Counter(handler_names)
        assert set(name_counts.values()) == {1}

    # Verify that once a handler fails, it is never re-executed again.
    handler_names = re.findall(r"'(.+?)' failed (?:permanently|with an exception. Will stop.)", runner.stdout)
    if e2e_failure_counts is not None:
        checked_names = [name for name in handler_names if name in e2e_failure_counts]
        name_counts = collections.Counter(checked_names)
        assert name_counts == e2e_failure_counts
    else:
        name_counts = collections.Counter(handler_names)
        assert not name_counts


def _parse_e2e_value(path: str, name: str) -> Any:
    with open(path, 'rt', encoding='utf-8') as f:
        name = re.escape(name)
        text = f.read()
        m = re.search(fr'^{name}\s*=\s*(.+)$', text, re.M)
        return eval(m.group(1)) if m else None


def _parse_e2e_presence(path: str, pattern: str) -> bool:
    with open(path, 'rt', encoding='utf-8') as f:
        text = f.read()
        m = re.search(pattern, text, re.M)
        return bool(m)


def _sleep_till_stopword(
        caplog,
        delay: float,
        patterns: Sequence[str] = (),
        *,
        interval: Optional[float] = None,
) -> bool:
    patterns = list(patterns or [])
    delay = delay or (10.0 if patterns else 1.0)
    interval = interval or min(1.0, max(0.1, delay / 10.))
    started = time.perf_counter()
    found = False
    while not found and time.perf_counter() - started < delay:
        for message in list(caplog.messages):
            if any(re.search(pattern, message) for pattern in patterns):
                found = True
                break
        else:
            time.sleep(interval)
    return found



import logging

import pytest

from kopf.reactor.causation import ActivityCause, DaemonCause, ResourceChangingCause, \
                                   ResourceSpawningCause, ResourceWatchingCause
from kopf.reactor.indexing import OperatorIndexer, OperatorIndexers
from kopf.reactor.invocation import build_kwargs
from kopf.structs.bodies import Body, BodyEssence
from kopf.structs.configuration import OperatorSettings
from kopf.structs.diffs import Diff
from kopf.structs.ephemera import Memo
from kopf.structs.handlers import Activity, Reason
from kopf.structs.patches import Patch
from kopf.structs.primitives import DaemonStopper


@pytest.fixture()
def indices():
    indexers = OperatorIndexers()
    indexers['index1'] = OperatorIndexer()
    indexers['index2'] = OperatorIndexer()
    return indexers.indices


@pytest.mark.parametrize('activity', set(Activity) - {Activity.STARTUP})
def test_activity_kwargs(resource, activity, indices):
    cause = ActivityCause(
        memo=Memo(),
        logger=logging.getLogger('kopf.test.fake.logger'),
        indices=indices,
        activity=activity,
        settings=OperatorSettings(),
    )
    kwargs = build_kwargs(cause=cause, extrakwarg=123)
    assert set(kwargs) == {'extrakwarg', 'memo', 'logger', 'index1', 'index2', 'activity'}
    assert kwargs['extrakwarg'] == 123
    assert kwargs['index1'] is indices['index1']
    assert kwargs['index2'] is indices['index2']
    assert kwargs['logger'] is cause.logger
    assert kwargs['activity'] is activity


@pytest.mark.parametrize('activity', {Activity.STARTUP})
def test_startup_kwargs(resource, activity, indices):
    cause = ActivityCause(
        memo=Memo(),
        logger=logging.getLogger('kopf.test.fake.logger'),
        indices=indices,
        activity=activity,
        settings=OperatorSettings(),
    )
    kwargs = build_kwargs(cause=cause, extrakwarg=123)
    assert set(kwargs) == {'extrakwarg', 'memo', 'logger', 'index1', 'index2',
                           'activity', 'settings'}
    assert kwargs['extrakwarg'] == 123
    assert kwargs['index1'] is indices['index1']
    assert kwargs['index2'] is indices['index2']
    assert kwargs['logger'] is cause.logger
    assert kwargs['activity'] is activity
    assert kwargs['settings'] is cause.settings


def test_resource_watching_kwargs(resource, indices):
    body = {'metadata': {'uid': 'uid1', 'name': 'name1', 'namespace': 'ns1',
                         'labels': {'l1': 'v1'}, 'annotations': {'a1': 'v1'}},
            'spec': {'field': 'value'},
            'status': {'info': 'payload'}}
    cause = ResourceWatchingCause(
        logger=logging.getLogger('kopf.test.fake.logger'),
        indices=indices,
        resource=resource,
        patch=Patch(),
        memo=Memo(),
        body=Body(body),
        type='ADDED',
        raw={'type': 'ADDED', 'object': {}},
    )
    kwargs = build_kwargs(cause=cause, extrakwarg=123)
    assert set(kwargs) == {'extrakwarg', 'logger', 'index1', 'index2', 'resource',
                           'patch', 'event', 'type', 'memo',
                           'body', 'spec', 'status', 'meta', 'uid', 'name', 'namespace',
                           'labels', 'annotations'}
    assert kwargs['extrakwarg'] == 123
    assert kwargs['resource'] is cause.resource
    assert kwargs['index1'] is indices['index1']
    assert kwargs['index2'] is indices['index2']
    assert kwargs['logger'] is cause.logger
    assert kwargs['patch'] is cause.patch
    assert kwargs['event'] is cause.raw
    assert kwargs['memo'] is cause.memo
    assert kwargs['type'] is cause.type
    assert kwargs['body'] is cause.body
    assert kwargs['spec'] is cause.body.spec
    assert kwargs['meta'] is cause.body.metadata
    assert kwargs['status'] is cause.body.status
    assert kwargs['labels'] is cause.body.metadata.labels
    assert kwargs['annotations'] is cause.body.metadata.annotations
    assert kwargs['uid'] == cause.body.metadata.uid
    assert kwargs['name'] == cause.body.metadata.name
    assert kwargs['namespace'] == cause.body.metadata.namespace


def test_resource_changing_kwargs(resource, indices):
    body = {'metadata': {'uid': 'uid1', 'name': 'name1', 'namespace': 'ns1',
                         'labels': {'l1': 'v1'}, 'annotations': {'a1': 'v1'}},
            'spec': {'field': 'value'},
            'status': {'info': 'payload'}}
    cause = ResourceChangingCause(
        logger=logging.getLogger('kopf.test.fake.logger'),
        indices=indices,
        resource=resource,
        patch=Patch(),
        initial=False,
        reason=Reason.NOOP,
        memo=Memo(),
        body=Body(body),
        diff=Diff([]),
        old=BodyEssence(),
        new=BodyEssence(),
    )
    kwargs = build_kwargs(cause=cause, extrakwarg=123)
    assert set(kwargs) == {'extrakwarg', 'logger', 'index1', 'index2', 'resource',
                           'patch', 'reason', 'memo',
                           'body', 'spec', 'status', 'meta', 'uid', 'name', 'namespace',
                           'labels', 'annotations', 'diff', 'old', 'new'}
    assert kwargs['extrakwarg'] == 123
    assert kwargs['resource'] is cause.resource
    assert kwargs['index1'] is indices['index1']
    assert kwargs['index2'] is indices['index2']
    assert kwargs['reason'] is cause.reason
    assert kwargs['logger'] is cause.logger
    assert kwargs['patch'] is cause.patch
    assert kwargs['memo'] is cause.memo
    assert kwargs['diff'] is cause.diff
    assert kwargs['old'] is cause.old
    assert kwargs['new'] is cause.new
    assert kwargs['body'] is cause.body
    assert kwargs['spec'] is cause.body.spec
    assert kwargs['meta'] is cause.body.metadata
    assert kwargs['status'] is cause.body.status
    assert kwargs['labels'] is cause.body.metadata.labels
    assert kwargs['annotations'] is cause.body.metadata.annotations
    assert kwargs['uid'] == cause.body.metadata.uid
    assert kwargs['name'] == cause.body.metadata.name
    assert kwargs['namespace'] == cause.body.metadata.namespace


def test_resource_spawning_kwargs(resource, indices):
    body = {'metadata': {'uid': 'uid1', 'name': 'name1', 'namespace': 'ns1',
                         'labels': {'l1': 'v1'}, 'annotations': {'a1': 'v1'}},
            'spec': {'field': 'value'},
            'status': {'info': 'payload'}}
    cause = ResourceSpawningCause(
        logger=logging.getLogger('kopf.test.fake.logger'),
        indices=indices,
        resource=resource,
        patch=Patch(),
        memo=Memo(),
        body=Body(body),
        reset=False,
    )
    kwargs = build_kwargs(cause=cause, extrakwarg=123)
    assert set(kwargs) == {'extrakwarg', 'logger', 'index1', 'index2',
                           'resource', 'patch', 'memo',
                           'body', 'spec', 'status', 'meta', 'uid', 'name', 'namespace',
                           'labels', 'annotations'}
    assert kwargs['extrakwarg'] == 123
    assert kwargs['resource'] is cause.resource
    assert kwargs['index1'] is indices['index1']
    assert kwargs['index2'] is indices['index2']
    assert kwargs['logger'] is cause.logger
    assert kwargs['patch'] is cause.patch
    assert kwargs['memo'] is cause.memo
    assert kwargs['body'] is cause.body
    assert kwargs['spec'] is cause.body.spec
    assert kwargs['meta'] is cause.body.metadata
    assert kwargs['status'] is cause.body.status
    assert kwargs['labels'] is cause.body.metadata.labels
    assert kwargs['annotations'] is cause.body.metadata.annotations
    assert kwargs['uid'] == cause.body.metadata.uid
    assert kwargs['name'] == cause.body.metadata.name
    assert kwargs['namespace'] == cause.body.metadata.namespace


def test_daemon_kwargs(resource, indices):
    body = {'metadata': {'uid': 'uid1', 'name': 'name1', 'namespace': 'ns1',
                         'labels': {'l1': 'v1'}, 'annotations': {'a1': 'v1'}},
            'spec': {'field': 'value'},
            'status': {'info': 'payload'}}
    cause = DaemonCause(
        logger=logging.getLogger('kopf.test.fake.logger'),
        indices=indices,
        resource=resource,
        patch=Patch(),
        memo=Memo(),
        body=Body(body),
        stopper=DaemonStopper(),
    )
    kwargs = build_kwargs(cause=cause, extrakwarg=123)
    assert set(kwargs) == {'extrakwarg', 'logger', 'index1', 'index2',
                           'resource', 'patch', 'memo',
                           'body', 'spec', 'status', 'meta', 'uid', 'name', 'namespace',
                           'labels', 'annotations'}
    assert kwargs['extrakwarg'] == 123
    assert kwargs['resource'] is cause.resource
    assert kwargs['index1'] is indices['index1']
    assert kwargs['index2'] is indices['index2']
    assert kwargs['logger'] is cause.logger
    assert kwargs['patch'] is cause.patch
    assert kwargs['memo'] is cause.memo
    assert kwargs['body'] is cause.body
    assert kwargs['spec'] is cause.body.spec
    assert kwargs['meta'] is cause.body.metadata
    assert kwargs['status'] is cause.body.status
    assert kwargs['labels'] is cause.body.metadata.labels
    assert kwargs['annotations'] is cause.body.metadata.annotations
    assert kwargs['uid'] == cause.body.metadata.uid
    assert kwargs['name'] == cause.body.metadata.name
    assert kwargs['namespace'] == cause.body.metadata.namespace
    assert 'stopped' not in kwargs


def test_daemon_sync_stopper(resource, indices):
    cause = DaemonCause(
        logger=logging.getLogger('kopf.test.fake.logger'),
        indices=indices,
        resource=resource,
        patch=Patch(),
        memo=Memo(),
        body=Body({}),
        stopper=DaemonStopper(),
    )
    kwargs = build_kwargs(cause=cause, _sync=True)
    assert kwargs['stopped'] is cause.stopper.sync_checker


def test_daemon_async_stopper(resource, indices):
    cause = DaemonCause(
        logger=logging.getLogger('kopf.test.fake.logger'),
        indices=indices,
        resource=resource,
        patch=Patch(),
        memo=Memo(),
        body=Body({}),
        stopper=DaemonStopper(),
    )
    kwargs = build_kwargs(cause=cause, _sync=False)
    assert kwargs['stopped'] is cause.stopper.async_checker



import asyncio
import logging

import pytest

import kopf
from kopf.reactor.indexing import OperatorIndexers
from kopf.reactor.processing import process_resource_event
from kopf.structs.containers import ResourceMemories
from kopf.structs.ephemera import Memo
from kopf.structs.handlers import HANDLER_REASONS, ResourceChangingHandler

LAST_SEEN_ANNOTATION = 'kopf.zalando.org/last-handled-configuration'


@pytest.mark.parametrize('cause_type', HANDLER_REASONS)
async def test_skipped_with_no_handlers(
        registry, settings, selector, resource, cause_mock, cause_type,
        caplog, assert_logs, k8s_mocked):
    caplog.set_level(logging.DEBUG)

    event_type = None
    event_body = {'metadata': {'finalizers': []}}
    cause_mock.reason = cause_type

    assert not registry._resource_changing.has_handlers(resource=resource)  # prerequisite
    registry._resource_changing.append(ResourceChangingHandler(
        reason='a-non-existent-cause-type',
        fn=lambda **_: None, id='id', param=None,
        errors=None, timeout=None, retries=None, backoff=None,
        selector=selector, annotations=None, labels=None, when=None,
        field=None, value=None, old=None, new=None, field_needs_change=None,
        deleted=None, initial=None, requires_finalizer=None,
    ))

    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': event_body},
        event_queue=asyncio.Queue(),
    )

    assert not k8s_mocked.sleep_or_wait.called
    assert k8s_mocked.patch_obj.called

    # The patch must contain ONLY the last-seen update, and nothing else.
    patch = k8s_mocked.patch_obj.call_args_list[0][1]['patch']
    assert set(patch.keys()) == {'metadata'}
    assert set(patch['metadata'].keys()) == {'annotations'}
    assert set(patch['metadata']['annotations'].keys()) == {LAST_SEEN_ANNOTATION}

    assert_logs([
        "(Creation|Updating|Resuming|Deletion) is in progress:",
        "Patching with:",
    ], prohibited=[
        "(Creation|Updating|Resuming|Deletion) is processed:",
    ])


@pytest.mark.parametrize('cause_type', HANDLER_REASONS)
@pytest.mark.parametrize('initial', [True, False, None])
@pytest.mark.parametrize('deleted', [True, False, None])
@pytest.mark.parametrize('annotations, labels, when', [
    (None, {'some-label': '...'}, None),
    ({'some-annotation': '...'}, None, None),
    (None, None, lambda **_: False),
])
async def test_stealth_mode_with_mismatching_handlers(
        registry, settings, selector, resource, cause_mock, cause_type,
        caplog, assert_logs, k8s_mocked, annotations, labels, when, deleted, initial):
    caplog.set_level(logging.DEBUG)

    event_type = None
    event_body = {'metadata': {'finalizers': []}}
    cause_mock.reason = cause_type

    assert not registry._resource_changing.has_handlers(resource=resource)  # prerequisite
    registry._resource_changing.append(ResourceChangingHandler(
        reason=None,
        fn=lambda **_: None, id='id', param=None,
        errors=None, timeout=None, retries=None, backoff=None,
        selector=selector, annotations=annotations, labels=labels, when=when,
        field=None, value=None, old=None, new=None, field_needs_change=None,
        deleted=deleted, initial=initial, requires_finalizer=None,
    ))

    await process_resource_event(
        lifecycle=kopf.lifecycles.all_at_once,
        registry=registry,
        settings=settings,
        resource=resource,
        indexers=OperatorIndexers(),
        memories=ResourceMemories(),
        memobase=Memo(),
        raw_event={'type': event_type, 'object': event_body},
        event_queue=asyncio.Queue(),
    )

    assert not k8s_mocked.sleep_or_wait.called
    assert not k8s_mocked.patch_obj.called
    assert not caplog.messages  # total stealth mode!



from kopf.storage.states import HandlerOutcome
from kopf.structs.callbacks import Result


def test_creation_for_ignored_handlers():
    outcome = HandlerOutcome(final=True)
    assert outcome.final
    assert outcome.delay is None
    assert outcome.result is None
    assert outcome.exception is None
    assert not outcome.subrefs


def test_creation_for_results():
    result = Result(object())
    outcome = HandlerOutcome(final=True, result=result)
    assert outcome.final
    assert outcome.delay is None
    assert outcome.result is result
    assert outcome.exception is None
    assert not outcome.subrefs


def test_creation_for_permanent_errors():
    error = Exception()
    outcome = HandlerOutcome(final=True, exception=error)
    assert outcome.final
    assert outcome.delay is None
    assert outcome.result is None
    assert outcome.exception is error
    assert not outcome.subrefs


def test_creation_for_temporary_errors():
    error = Exception()
    outcome = HandlerOutcome(final=False, exception=error, delay=123)
    assert not outcome.final
    assert outcome.delay == 123
    assert outcome.result is None
    assert outcome.exception is error
    assert not outcome.subrefs


def test_creation_with_subrefs():
    outcome = HandlerOutcome(final=True, subrefs=['sub1', 'sub2'])
    assert outcome.subrefs == ['sub1', 'sub2']



import pytest

from kopf.reactor.activities import authenticate
from kopf.reactor.indexing import OperatorIndexers
from kopf.reactor.registries import OperatorRegistry
from kopf.structs.credentials import ConnectionInfo, LoginError, Vault
from kopf.structs.ephemera import Memo
from kopf.structs.handlers import Activity, ActivityHandler


async def test_empty_registry_produces_no_credentials(settings):
    vault = Vault()
    registry = OperatorRegistry()

    await authenticate(
        registry=registry,
        settings=settings,
        vault=vault,
        memo=Memo(),
        indices=OperatorIndexers().indices,
    )

    assert not vault
    with pytest.raises(LoginError):
        async for _, _ in vault:
            pass


async def test_noreturn_handler_produces_no_credentials(settings):
    vault = Vault()
    registry = OperatorRegistry()

    def login_fn(**_):
        pass

    # NB: id auto-detection does not work, as it is local to the test function.
    registry._activities.append(ActivityHandler(
        fn=login_fn, id='login_fn', activity=Activity.AUTHENTICATION,
        param=None, errors=None, timeout=None, retries=None, backoff=None,
    ))

    await authenticate(
        registry=registry,
        settings=settings,
        vault=vault,
        memo=Memo(),
        indices=OperatorIndexers().indices,
    )

    assert not vault
    with pytest.raises(LoginError):
        async for _, _ in vault:
            pass


async def test_single_credentials_provided_to_vault(settings):
    info = ConnectionInfo(server='https://expected/')
    vault = Vault()
    registry = OperatorRegistry()

    def login_fn(**_):
        return info

    # NB: id auto-detection does not work, as it is local to the test function.
    registry._activities.append(ActivityHandler(
        fn=login_fn, id='login_fn', activity=Activity.AUTHENTICATION,
        param=None, errors=None, timeout=None, retries=None, backoff=None,
    ))

    await authenticate(
        registry=registry,
        settings=settings,
        vault=vault,
        memo=Memo(),
        indices=OperatorIndexers().indices,
    )

    assert vault

    items = []
    async for key, info in vault:
        items.append((key, info))

    assert len(items) == 1
    assert items[0][0] == 'login_fn'
    assert items[0][1] is info



"""
Embeddable operators require very customised application-specific testing.
Kopf cannot help here beyond its regular `kopf.testing.KopfRunner` helper,
which is an equivalent of `kopf run` command.

This file exists to disable the implicit e2e tests
(they skip if explicit e2e tests exist in the example directory).
"""



import pytest

from kopf.reactor.indexing import OperatorIndexer, OperatorIndexers
from kopf.structs.bodies import Body


@pytest.fixture()
def indexers():
    return OperatorIndexers()


@pytest.fixture()
def index(indexers):
    indexer = OperatorIndexer()
    indexers['index_fn'] = indexer
    return indexer.index


@pytest.fixture()
async def indexed_123(indexers, index):
    body = {'metadata': {'namespace': 'ns1', 'name': 'name1'}}
    key = indexers.make_key(Body(body))
    indexers['index_fn'].replace(key, 123)
    assert set(index) == {None}
    assert set(index[None]) == {123}



import pytest

from kopf.clients.errors import APIError
from kopf.clients.watching import Bookmark, infinite_watch

STREAM_WITH_UNKNOWN_EVENT = [
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'UNKNOWN', 'object': {}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
]
STREAM_WITH_ERROR_410GONE = [
    {'type': 'ADDED', 'object': {'spec': 'a'}},
    {'type': 'ERROR', 'object': {'code': 410}},
    {'type': 'ADDED', 'object': {'spec': 'b'}},
]


class SampleException(Exception):
    pass


async def test_exception_escalates(
        settings, resource, stream, namespace, enforced_session, mocker):

    enforced_session.get = mocker.Mock(side_effect=SampleException())
    stream.feed([], namespace=namespace)
    stream.close(namespace=namespace)

    events = []
    with pytest.raises(SampleException):
        async for event in infinite_watch(settings=settings,
                                          resource=resource,
                                          namespace=namespace,
                                          _iterations=1):
            events.append(event)

    assert len(events) == 0


async def test_infinite_watch_never_exits_normally(
        settings, resource, stream, namespace, aresponses):
    error = aresponses.Response(status=555)
    stream.feed(
        STREAM_WITH_ERROR_410GONE,  # watching restarted
        STREAM_WITH_UNKNOWN_EVENT,  # event ignored
        error,  # to finally exit it somehow
        namespace=namespace,
    )
    stream.close(namespace=namespace)

    events = []
    with pytest.raises(APIError) as e:
        async for event in infinite_watch(settings=settings,
                                          resource=resource,
                                          namespace=namespace):
            events.append(event)

    assert e.value.status == 555

    assert len(events) == 5
    assert events[0] == Bookmark.LISTED
    assert events[1]['object']['spec'] == 'a'
    assert events[2] == Bookmark.LISTED
    assert events[3]['object']['spec'] == 'a'
    assert events[4]['object']['spec'] == 'b'



import pytest

from kopf.engines.peering import guess_selector
from kopf.structs.references import CLUSTER_PEERINGS, NAMESPACED_PEERINGS


@pytest.mark.parametrize('namespaced, expected_resource', [
    (False, CLUSTER_PEERINGS),
    (True, NAMESPACED_PEERINGS),
])
@pytest.mark.parametrize('mandatory', [False, True])
def test_resource_when_not_standalone(settings, namespaced, mandatory, expected_resource):
    settings.peering.standalone = False
    settings.peering.namespaced = namespaced
    settings.peering.mandatory = mandatory
    selector = guess_selector(settings=settings)
    assert selector == expected_resource


@pytest.mark.parametrize('namespaced', [False, True])
@pytest.mark.parametrize('mandatory', [False, True])
def test_resource_when_standalone(settings, namespaced, mandatory):
    settings.peering.standalone = True
    settings.peering.namespaced = namespaced
    settings.peering.mandatory = mandatory
    selector = guess_selector(settings=settings)
    assert selector is None



import asyncio
import logging

import async_timeout
import pytest

from kopf.reactor.lifecycles import all_at_once
from kopf.reactor.processing import process_resource_event
from kopf.structs.containers import ResourceMemories
from kopf.structs.ephemera import Memo
from kopf.structs.primitives import ToggleSet

EVENT_TYPES_WHEN_EXISTS = [None, 'ADDED', 'MODIFIED']
EVENT_TYPES_WHEN_GONE = ['DELETED']
EVENT_TYPES = EVENT_TYPES_WHEN_EXISTS + EVENT_TYPES_WHEN_GONE


@pytest.mark.parametrize('event_type', EVENT_TYPES_WHEN_EXISTS)
async def test_reporting_on_resource_readiness(
        resource, settings, registry, indexers, caplog, event_type, handlers, timer):
    caplog.set_level(logging.DEBUG)

    operator_indexed = ToggleSet(all)
    resource_indexed = await operator_indexed.make_toggle()
    async with timer, async_timeout.timeout(0.5) as timeout:
        await process_resource_event(
            lifecycle=all_at_once,
            registry=registry,
            settings=settings,
            resource=resource,
            indexers=indexers,
            memories=ResourceMemories(),
            memobase=Memo(),
            raw_event={'type': event_type, 'object': {}},
            event_queue=asyncio.Queue(),
            operator_indexed=operator_indexed,
            resource_indexed=resource_indexed,
        )
    assert not timeout.expired
    assert timer.seconds < 0.2  # asap, nowait
    assert operator_indexed.is_on()
    assert set(operator_indexed) == set()  # save RAM
    assert handlers.event_mock.called


@pytest.mark.parametrize('event_type', EVENT_TYPES_WHEN_EXISTS)
async def test_blocking_when_operator_is_not_ready(
        resource, settings, registry, indexers, caplog, event_type, handlers, timer):
    caplog.set_level(logging.DEBUG)

    operator_indexed = ToggleSet(all)
    resource_listed = await operator_indexed.make_toggle()
    resource_indexed = await operator_indexed.make_toggle()
    with pytest.raises(asyncio.TimeoutError):
        async with timer, async_timeout.timeout(0.2) as timeout:
            await process_resource_event(
                lifecycle=all_at_once,
                registry=registry,
                settings=settings,
                resource=resource,
                indexers=indexers,
                memories=ResourceMemories(),
                memobase=Memo(),
                raw_event={'type': event_type, 'object': {}},
                event_queue=asyncio.Queue(),
                operator_indexed=operator_indexed,
                resource_indexed=resource_indexed,
            )
    assert timeout.expired
    assert 0.2 < timer.seconds < 0.4
    assert operator_indexed.is_off()
    assert set(operator_indexed) == {resource_listed}
    assert not handlers.event_mock.called


@pytest.mark.parametrize('event_type', EVENT_TYPES_WHEN_EXISTS)
async def test_unblocking_once_operator_is_ready(
        resource, settings, registry, indexers, caplog, event_type, handlers, timer):
    caplog.set_level(logging.DEBUG)

    async def delayed_readiness(delay: float):
        await asyncio.sleep(delay)
        await resource_listed.turn_to(True)

    operator_indexed = ToggleSet(all)
    resource_listed = await operator_indexed.make_toggle()
    resource_indexed = await operator_indexed.make_toggle()
    async with timer, async_timeout.timeout(1.0) as timeout:
        asyncio.create_task(delayed_readiness(0.2))
        await process_resource_event(
            lifecycle=all_at_once,
            registry=registry,
            settings=settings,
            resource=resource,
            indexers=indexers,
            memories=ResourceMemories(),
            memobase=Memo(),
            raw_event={'type': event_type, 'object': {}},
            event_queue=asyncio.Queue(),
            operator_indexed=operator_indexed,
            resource_indexed=resource_indexed,
        )
    assert not timeout.expired
    assert 0.2 < timer.seconds < 0.4
    assert operator_indexed.is_on()
    assert set(operator_indexed) == {resource_listed}
    assert handlers.event_mock.called



import asyncio
import functools

import pytest
from asynctest import CoroutineMock

from kopf.clients.watching import infinite_watch
from kopf.reactor.queueing import watcher, worker as original_worker


@pytest.fixture(autouse=True)
def _autouse_resp_mocker(resp_mocker):
    pass


@pytest.fixture()
def processor():
    """ A mock for processor -- to be checked if the handler has been called. """
    return CoroutineMock()


@pytest.fixture()
def worker_spy(mocker):
    """ Spy on the watcher: actually call it, but provide the mock-fields. """
    spy = CoroutineMock(spec=original_worker, wraps=original_worker)
    return mocker.patch('kopf.reactor.queueing.worker', spy)


@pytest.fixture()
def worker_mock(mocker):
    """ Prevent the queue consumption, so that the queues could be checked. """
    return mocker.patch('kopf.reactor.queueing.worker')


@pytest.fixture()
def watcher_limited(mocker, settings):
    """ Make event streaming finite, watcher exits after depletion. """
    settings.watching.reconnect_backoff = 0
    mocker.patch('kopf.clients.watching.infinite_watch',
                 new=functools.partial(infinite_watch, _iterations=1))


@pytest.fixture()
def watcher_in_background(settings, resource, event_loop, worker_spy, stream):

    # Prevent remembering the streaming objects in the mocks.
    async def do_nothing(*args, **kwargs):
        pass

    # Prevent any real streaming for the very beginning, before it even starts.
    stream.feed([])

    # Spawn a watcher in the background.
    coro = watcher(
        namespace=None,
        resource=resource,
        settings=settings,
        processor=do_nothing,
    )
    task = event_loop.create_task(coro)

    try:
        # Go for a test.
        yield task
    finally:
        # Terminate the watcher to cleanup the loop.
        task.cancel()
        try:
            event_loop.run_until_complete(task)
        except asyncio.CancelledError:
            pass



import aiohttp
import pytest

from kopf.clients.auth import APIContext, reauthenticated_request
from kopf.clients.errors import APIError, APIForbiddenError, APINotFoundError, \
                                APIUnauthorizedError, check_response


@reauthenticated_request
async def get_it(url: str, *, context: APIContext) -> None:
    response = await context.session.get(url)
    await check_response(response)
    return await response.json()


def test_aiohttp_is_not_leaked_outside():
    assert not issubclass(APIError, aiohttp.ClientError)


def test_exception_without_payload():
    exc = APIError(None, status=456)
    assert exc.status == 456
    assert exc.code is None
    assert exc.message is None
    assert exc.details is None


def test_exception_with_payload():
    exc = APIError({"message": "msg", "code": 123, "details": {"a": "b"}}, status=456)
    assert exc.status == 456
    assert exc.code == 123
    assert exc.message == "msg"
    assert exc.details == {"a": "b"}


@pytest.mark.parametrize('status', [200, 202, 300, 304])
async def test_no_error_on_success(
        resp_mocker, aresponses, hostname, status):

    resp = aresponses.Response(
        status=status,
        headers={'Content-Type': 'application/json'},
        text='{"kind": "Status", "code": "xxx", "message": "msg"}',
    )
    aresponses.add(hostname, '/', 'get', resp_mocker(return_value=resp))

    await get_it(f"http://{hostname}/")


@pytest.mark.parametrize('status, exctype', [
    (400, APIError),
    (401, APIUnauthorizedError),
    (403, APIForbiddenError),
    (404, APINotFoundError),
    (500, APIError),
    (666, APIError),
])
async def test_error_with_payload(
        resp_mocker, aresponses, hostname, status, exctype):

    resp = aresponses.Response(
        status=status,
        headers={'Content-Type': 'application/json'},
        text='{"kind": "Status", "code": 123, "message": "msg", "details": {"a": "b"}}',
    )
    aresponses.add(hostname, '/', 'get', resp_mocker(return_value=resp))

    with pytest.raises(APIError) as err:
        await get_it(f"http://{hostname}/")

    assert not isinstance(err.value, aiohttp.ClientResponseError)
    assert isinstance(err.value, exctype)
    assert err.value.status == status
    assert err.value.code == 123
    assert err.value.message == 'msg'
    assert err.value.details == {'a': 'b'}


@pytest.mark.parametrize('status', [400, 500, 666])
async def test_error_with_nonjson_payload(
        resp_mocker, aresponses, hostname, status):

    resp = aresponses.Response(
        status=status,
        headers={'Content-Type': 'application/json'},
        text='unparsable json',
    )
    aresponses.add(hostname, '/', 'get', resp_mocker(return_value=resp))

    with pytest.raises(APIError) as err:
        await get_it(f"http://{hostname}/")

    assert err.value.status == status
    assert err.value.code is None
    assert err.value.message is None
    assert err.value.details is None


@pytest.mark.parametrize('status', [400, 500, 666])
async def test_error_with_parseable_nonk8s_payload(
        resp_mocker, aresponses, hostname, status):

    resp = aresponses.Response(
        status=status,
        headers={'Content-Type': 'application/json'},
        text='{"kind": "NonStatus", "code": "xxx", "message": "msg"}',
    )
    aresponses.add(hostname, '/', 'get', resp_mocker(return_value=resp))

    with pytest.raises(APIError) as err:
        await get_it(f"http://{hostname}/")

    assert err.value.status == status
    assert err.value.code is None
    assert err.value.message is None
    assert err.value.details is None



import pytest


@pytest.mark.parametrize('kwarg, value, options, envvars', [
    ('paths', (), [], {}),
    ('paths', ('path1', 'path2'), ['path1', 'path2'], {}),
    # ('paths', ('path1', 'path2'), [], {'KOPF_RUN_PATHS': 'path1 path2'}), FIXME: UNSUPPORTED

    ('modules', (), [], {}),
    ('modules', ('mod1', 'mod2'), ['-m', 'mod1', '-m', 'mod2'], {}),
    ('modules', ('mod1', 'mod2'), ['--module', 'mod1', '--module', 'mod2'], {}),
    ('modules', ('mod1', 'mod2'), [], {'KOPF_RUN_MODULES': 'mod1 mod2'}),
], ids=[
    'default-paths', 'arg-paths', #'env-paths',
    'default-modules', 'opt-short-m', 'opt-long-modules', 'env-modules',
])
def test_options_passed_to_preload(invoke, options, envvars, kwarg, value, preload, real_run):
    result = invoke(['run'] + options, env=envvars)
    assert result.exit_code == 0
    assert preload.called
    assert preload.call_args[1][kwarg] == value


@pytest.mark.parametrize('kwarg, value, options, envvars', [
    ('standalone', None, [], {}),
    ('standalone', True, ['--standalone'], {}),
    ('standalone', True, [], {'KOPF_RUN_STANDALONE': 'true'}),

    ('namespaces', (), [], {}),
    ('namespaces', ('ns',), ['-n', 'ns'], {}),
    ('namespaces', ('ns',), ['--namespace=ns'], {}),
    ('namespaces', ('ns',), [], {'KOPF_RUN_NAMESPACE': 'ns'}),
    ('namespaces', ('ns',), [], {'KOPF_RUN_NAMESPACES': 'ns'}),

    ('namespaces', ('ns1', 'ns2'), ['--namespace=ns1', '-n', 'ns2'], {}),
    ('namespaces', ('ns1', 'ns2'), [], {'KOPF_RUN_NAMESPACES': 'ns1 ns2'}),

    ('peering_name', None, [], {}),
    ('peering_name', 'peer', ['-P', 'peer'], {}),
    ('peering_name', 'peer', ['--peering=peer'], {}),
    ('peering_name', 'peer', [], {'KOPF_RUN_PEERING': 'peer'}),

    ('priority', None, [], {}),
    ('priority', 123, ['-p', '123'], {}),
    ('priority', 123, ['--priority=123'], {}),
    ('priority', 123, [], {'KOPF_RUN_PRIORITY': '123'}),
    ('priority', 666, ['--dev'], {}),
], ids=[
    'default-standalone', 'opt-long-standalone', 'env-standalone',
    'default-namespace', 'opt-short-n', 'opt-long-namespace', 'env1-namespace', 'env2-namespace',
    'opt-multi-namespaces', 'env-multi-namespaces',
    'default-peering', 'opt-short-P', 'opt-long-peering', 'env-peering',
    'default-priority', 'opt-short-p', 'opt-long-priority', 'env-priority', 'opt-long-dev',
])
def test_options_passed_to_realrun(invoke, options, envvars, kwarg, value, preload, real_run):
    result = invoke(['run'] + options, env=envvars)
    assert result.exit_code == 0
    assert real_run.called
    assert real_run.call_args[1][kwarg] == value



import pprint

import kopf


@kopf.index('pods')
def is_running(namespace, name, status, **_):
    return {(namespace, name): status.get('phase') == 'Running'}
    # {('kube-system', 'traefik-...-...'): True,
    #  ('kube-system', 'helm-install-traefik-...'): False,
    #    ...}


@kopf.index('pods')
def by_label(labels, name, **_):
    return {(label, value): name for label, value in labels.items()}
    # {('app', 'traefik'): ['traefik-...-...'],
    #  ('job-name', 'helm-install-traefik'): ['helm-install-traefik-...'],
    #  ('helmcharts.helm.cattle.io/chart', 'traefik'): ['helm-install-traefik-...'],
    #    ...}


@kopf.on.probe()
def pod_count(is_running: kopf.Index, **_):
    return len(is_running)


@kopf.on.probe()
def pod_names(is_running: kopf.Index, **_):
    return [name for _, name in is_running]


@kopf.timer('kex', interval=5)
def intervalled(is_running: kopf.Index, by_label: kopf.Index, patch: kopf.Patch, **_):
    pprint.pprint(dict(by_label))
    patch.status['running-pods'] = [
        f"{ns}::{name}"
        for (ns, name), is_running in is_running.items()
        if ns in ['kube-system', 'default']
        if is_running
    ]


# Marks for the e2e tests (see tests/e2e/test_examples.py):
E2E_SUCCESS_COUNTS = {}  # we do not care: pods can have 6-10 updates here.



import asyncio
import dataclasses
import enum
import fnmatch
import re
import urllib.parse
from typing import Collection, FrozenSet, Iterable, Iterator, List, Mapping, \
                   MutableMapping, NewType, Optional, Pattern, Set, Union

# A namespace specification with globs, negations, and some minimal syntax; see `match_namespace()`.
# Regexps are also supported if pre-compiled from the code, not from the CLI options as raw strings.
NamespacePattern = Union[str, Pattern]

# A specific really existing addressable namespace (at least, the one assumed to be so).
# Made as a NewType for stricter type-checking to avoid collisions with patterns and other strings.
NamespaceName = NewType('NamespaceName', str)

# A namespace reference usable in the API calls. `None` means cluster-wide API calls.
Namespace = Optional[NamespaceName]


def select_specific_namespaces(patterns: Iterable[NamespacePattern]) -> Collection[NamespaceName]:
    """
    Select the namespace specifications that can be used as direct namespaces.

    It is used in a fallback scenario when the namespace observation is either
    disabled or not possible due to restricted permission, while the normal
    operation is still possible in the very specific configured namespaces.
    """
    return {
        NamespaceName(pattern)
        for pattern in patterns
        if isinstance(pattern, str)  # excl. regexps & etc.
        if not('!' in pattern or '*' in pattern or '?' in pattern or ',' in pattern)
    }


def match_namespace(name: NamespaceName, pattern: NamespacePattern) -> bool:
    """
    Check if the specific namespace matches a namespace specification.

    Each individual namespace pattern is a string that follows some syntax:

    * the pattern consists of comma-separated parts (spaces are ignored);
    * each part is either an inclusive or an exclusive (negating) glob;
    * each glob can have ``*`` and ``?`` placeholders for any or one symbols;
    * the exclusive globs start with ``!``;
    * if the the first glob is exclusive, then a preceding catch-all is implied.

    A check of whether a namespace matches the individual pattern, is done by
    iterating the pattern's globs left-to-right: the exclusive patterns exclude
    it from the match; the first inclusive pattern does the initial match, while
    the following inclusive patterns only re-match it if it was excluded before;
    i.e., they do not do the full initial match.

    For example, the pattern ``"myapp-*, !*-pr-*, *pr-123"``
    will match ``myapp-test``, ``myapp-live``, even ``myapp-pr-123``,
    but not ``myapp-pr-456`` and certainly not ``otherapp-pr-123``.
    The latter one, despite it matches the last glob, is not included
    because it was not matched by the initial pattern.

    On the other hand, the pattern ``"!*-pr-*, *pr-123"``
    (equivalent to ``"*, !*-pr-*, *pr-123"``) will match ``myapp-test``,
    ``myapp-live``, ``myapp-pr-123``, ``anyapp-anything``,
    and even ``otherapp-pr-123`` -- though not ``myapp-pr-456``.
    Unlike in the first example, the otherapp's namespace was included initially
    by the first glob (the implied ``*``), and therefore could be re-matched
    by the last glob ``*pr-123`` after being excluded by ``!*-pr-*``.

    While these are theoretical capabilities of this pattern-matching algorithm,
    it is not expected that they will be abused too much. The main intention is
    to have simple one-glob patterns (either inclusive or exclusive),
    only rarely followed by a single negation.
    """

    # Regexps are powerful enough on their own -- we do not parse or interpret them.
    if isinstance(pattern, re.Pattern):
        return bool(pattern.fullmatch(name))

    # The first pattern should be an inclusive one. Unless it is, prepend a catch-all pattern.
    globs = [glob.strip() for glob in pattern.split(',')]
    if not globs or globs[0].startswith('!'):
        globs.insert(0, '*')

    # Iterate and calculate: every inclusive pattern makes the namespace to match regardless,
    # of the previous result; every exclusive pattern un-matches it if it was matched before.
    matches = first_match = fnmatch.fnmatch(name, globs[0])
    for glob in globs[1:]:
        if glob.startswith('!'):
            matches = matches and not fnmatch.fnmatch(name, glob.lstrip('!'))
        else:
            matches = matches or (first_match and fnmatch.fnmatch(name, glob))

    return matches


# Detect conventional API versions for some cases: e.g. in "myresources.v1alpha1.example.com".
# Non-conventional versions are indistinguishable from API groups ("myresources.foo1.example.com").
# See also: https://kubernetes.io/docs/tasks/extend-kubernetes/custom-resources/custom-resource-definition-versioning/
K8S_VERSION_PATTERN = re.compile(r'^v\d+(?:(?:alpha|beta)\d+)?$')


@dataclasses.dataclass(frozen=True, eq=False, repr=False)
class Resource:
    """
    A reference to a very specific custom or built-in resource kind.

    It is used to form the K8s API URLs. Generally, K8s API only needs
    an API group, an API version, and a plural name of the resource.
    All other names are remembered to match against resource selectors,
    for logging, and for informational purposes.
    """

    group: str
    """
    The resource's API group; e.g. ``"kopf.dev"``, ``"apps"``, ``"batch"``.
    For Core v1 API resources, an empty string: ``""``.
    """

    version: str
    """
    The resource's API version; e.g. ``"v1"``, ``"v1beta1"``, etc.
    """

    plural: str
    """
    The resource's plural name; e.g. ``"pods"``, ``"kopfexamples"``.
    It is used as an API endpoint, together with API group & version.
    """

    kind: Optional[str] = None
    """
    The resource's kind (as in YAML files); e.g. ``"Pod"``, ``"KopfExample"``.
    """

    singular: Optional[str] = None
    """
    The resource's singular name; e.g. ``"pod"``, ``"kopfexample"``.
    """

    shortcuts: FrozenSet[str] = frozenset()
    """
    The resource's short names; e.g. ``{"po"}``, ``{"kex", "kexes"}``.
    """

    categories: FrozenSet[str] = frozenset()
    """
    The resource's categories, to which the resource belongs; e.g. ``{"all"}``.
    """

    subresources: FrozenSet[str] = frozenset()
    """
    The resource's subresources, if defined; e.g. ``{"status", "scale"}``.
    """

    namespaced: Optional[bool] = None
    """
    Whether the resource is namespaced (``True``) or cluster-scoped (``False``).
    """

    preferred: bool = True  # against conventions, but makes versionless selectors match by default.
    """
    Whether the resource belong to a "preferred" API version.
    Only "preferred" resources are served when the version is not specified.
    """

    verbs: FrozenSet[str] = frozenset()
    """
    All available verbs for the resource, as supported by K8s API;
    e.g., ``{"list", "watch", "create", "update", "delete", "patch"}``.
    Note that it is not the same as all verbs permitted by RBAC.
    """

    def __hash__(self) -> int:
        return hash((self.group, self.version, self.plural))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Resource):
            self_tuple = (self.group, self.version, self.plural)
            other_tuple = (other.group, other.version, other.plural)
            return self_tuple == other_tuple
        else:
            return NotImplemented

    def __repr__(self) -> str:
        plural_main, *subs = self.plural.split('/')
        name_text = f'{plural_main}.{self.version}.{self.group}'.strip('.')
        subs_text = f'/{"/".join(subs)}' if subs else ''
        return f'{name_text}{subs_text}'

    # Mostly for tests, to be used as `@kopf.on.event(*resource, ...)`
    def __iter__(self) -> Iterator[str]:
        return iter((self.group, self.version, self.plural))

    def get_url(
            self,
            *,
            server: Optional[str] = None,
            namespace: Namespace = None,
            name: Optional[str] = None,
            subresource: Optional[str] = None,
            params: Optional[Mapping[str, str]] = None,
    ) -> str:
        """
        Build a URL to be used with K8s API.

        If the namespace is not set, a cluster-wide URL is returned.
        For cluster-scoped resources, the namespace is ignored.

        If the name is not set, the URL for the resource list is returned.
        Otherwise (if set), the URL for the individual resource is returned.

        If subresource is set, that subresource's URL is returned,
        regardless of whether such a subresource is known or not.

        Params go to the query parameters (``?param1=value1&param2=value2...``).
        """
        if subresource is not None and name is None:
            raise ValueError("Subresources can be used only with specific resources by their name.")
        if not self.namespaced and namespace is not None:
            raise ValueError(f"Specific namespaces are not supported for cluster-scoped resources.")
        if self.namespaced and namespace is None and name is not None:
            raise ValueError("Specific namespaces are required for specific namespaced resources.")

        parts: List[Optional[str]] = [
            '/api' if self.group == '' and self.version == 'v1' else '/apis',
            self.group,
            self.version,
            'namespaces' if self.namespaced and namespace is not None else None,
            namespace if self.namespaced and namespace is not None else None,
            self.plural,
            name,
            subresource,
        ]

        query = urllib.parse.urlencode(params, encoding='utf-8') if params else ''
        path = '/'.join([part for part in parts if part])
        url = path + ('?' if query else '') + query
        return url if server is None else server.rstrip('/') + '/' + url.lstrip('/')


class Marker(enum.Enum):
    """
    A special marker to handle all resources possible, built-in and custom.
    """
    EVERYTHING = enum.auto()


# An explicit catch-all marker for positional arguments of resource selectors.
EVERYTHING = Marker.EVERYTHING


@dataclasses.dataclass(frozen=True)
class Selector:
    """
    A resource specification that can match several resource kinds.

    The resource specifications are not usable in K8s API calls, as the API
    has no endpoints with masks or placeholders for unknown or catch-all
    resource identifying parts (e.g. any API group, any API version, any name).

    They are used only locally in the operator to match against the actual
    resources with specific names (:class:`Resource`). The handlers are
    defined with resource specifications, but are invoked with specific
    resource kinds. Even if those specifications look very concrete and allow
    no variations, they still remain specifications.
    """

    arg1: dataclasses.InitVar[Union[None, str, Marker]] = None
    arg2: dataclasses.InitVar[Union[None, str, Marker]] = None
    arg3: dataclasses.InitVar[Union[None, str, Marker]] = None
    argN: dataclasses.InitVar[None] = None  # a runtime guard against too many positional arguments

    group: Optional[str] = None
    version: Optional[str] = None

    kind: Optional[str] = None
    plural: Optional[str] = None
    singular: Optional[str] = None
    shortcut: Optional[str] = None
    category: Optional[str] = None
    any_name: Optional[Union[str, Marker]] = None

    def __post_init__(
            self,
            arg1: Union[None, str, Marker],
            arg2: Union[None, str, Marker],
            arg3: Union[None, str, Marker],
            argN: None,  # a runtime guard against too many positional arguments
    ) -> None:

        # Since the class is frozen & read-only, post-creation field adjustment is done via a hack.
        # This is the same hack as used in the frozen dataclasses to initialise their fields.
        if argN is not None:
            raise TypeError("Too many positional arguments. Max 3 positional args are accepted.")
        elif arg3 is not None:
            object.__setattr__(self, 'group', arg1)
            object.__setattr__(self, 'version', arg2)
            object.__setattr__(self, 'any_name', arg3)
        elif arg2 is not None and isinstance(arg1, str) and '/' in arg1:
            object.__setattr__(self, 'group', arg1.rsplit('/', 1)[0])
            object.__setattr__(self, 'version', arg1.rsplit('/')[-1])
            object.__setattr__(self, 'any_name', arg2)
        elif arg2 is not None and arg1 == 'v1':
            object.__setattr__(self, 'group', '')
            object.__setattr__(self, 'version', arg1)
            object.__setattr__(self, 'any_name', arg2)
        elif arg2 is not None:
            object.__setattr__(self, 'group', arg1)
            object.__setattr__(self, 'any_name', arg2)
        elif arg1 is not None and isinstance(arg1, Marker):
            object.__setattr__(self, 'any_name', arg1)
        elif arg1 is not None and '.' in arg1 and K8S_VERSION_PATTERN.match(arg1.split('.')[1]):
            if len(arg1.split('.')) >= 3:
                object.__setattr__(self, 'group', arg1.split('.', 2)[2])
            object.__setattr__(self, 'version', arg1.split('.')[1])
            object.__setattr__(self, 'any_name', arg1.split('.')[0])
        elif arg1 is not None and '.' in arg1:
            object.__setattr__(self, 'group', arg1.split('.', 1)[1])
            object.__setattr__(self, 'any_name', arg1.split('.')[0])
        elif arg1 is not None:
            object.__setattr__(self, 'any_name', arg1)

        # Verify that explicit & interpreted arguments have produced an unambiguous specification.
        names = [self.kind, self.plural, self.singular, self.shortcut, self.category, self.any_name]
        clean = [name for name in names if name is not None]
        if len(clean) > 1:
            raise TypeError(f"Ambiguous resource specification with names {clean}")
        if len(clean) < 1:
            raise TypeError(f"Unspecific resource with no names.")

        # For reasons unknown, the singular is empty for ALL builtin resources. This does not affect
        # the checks unless defined as e.g. ``singular=""``, which would match ALL builtins at once.
        # Thus we prohibit it until clarified why is it so, what does it mean, how to deal with it.
        if any([name == '' for name in names]):
            raise TypeError("Names must not be empty strings; either None or specific strings.")

    def __repr__(self) -> str:
        kwargs = {f.name: getattr(self, f.name) for f in dataclasses.fields(self)}
        kwtext = ', '.join([f'{key!s}={val!r}' for key, val in kwargs.items() if val is not None])
        clsname = self.__class__.__name__
        return f'{clsname}({kwtext})'

    @property
    def is_specific(self) -> bool:
        return (self.kind is not None or
                self.shortcut is not None or
                self.plural is not None or
                self.singular is not None or
                (self.any_name is not None and not isinstance(self.any_name, Marker)))

    def check(self, resource: Resource) -> bool:
        """
        Check if a specific resources matches this resource specification.
        """
        # Core v1 events are excluded from EVERYTHING: they are implicitly produced during handling,
        # and thus trigger unnecessary handling cycles (even for other resources, not for events).
        return (
            (self.group is None or self.group == resource.group) and
            ((self.version is None and resource.preferred) or self.version == resource.version) and
            (self.kind is None or self.kind == resource.kind) and
            (self.plural is None or self.plural == resource.plural) and
            (self.singular is None or self.singular == resource.singular) and
            (self.category is None or self.category in resource.categories) and
            (self.shortcut is None or self.shortcut in resource.shortcuts) and
            (self.any_name is None or
             self.any_name == resource.kind or
             self.any_name == resource.plural or
             self.any_name == resource.singular or
             self.any_name in resource.shortcuts or
             (self.any_name is Marker.EVERYTHING and
              not EVENTS.check(resource) and
              not EVENTS_K8S.check(resource))))

    def select(self, resources: Collection[Resource]) -> Collection[Resource]:
        result = {resource for resource in resources if self.check(resource)}

        # Core v1 API group's priority is hard-coded in K8s and kubectl. Do the same. For example:
        # whenever "pods" is specified, and "pods.v1" & "pods.v1beta1.metrics.k8s.io" are found,
        # implicitly give priority to "v1" and hide the existence of non-"v1" groups.
        # But not if they are specified by categories! -- In that case, keep all resources as is.
        if self.is_specific:
            v1only = {resource for resource in result if resource.group == ''}
            result = v1only or result

        return result


# Some predefined API endpoints that we use in the framework itself (not exposed to the operators).
# Note: the CRDs are versionless: we do not look into its ``spec`` stanza, we only watch for
# the fact of changes, so the schema does not matter, any cluster-preferred API version would work.
# Note: the peering resources are either zalando.org/v1 or kopf.dev/v1; both cannot co-exist because
# they would share the names, so K8s will not let this. It is done for domain name transitioning.
CRDS = Selector('apiextensions.k8s.io', 'customresourcedefinitions')
EVENTS = Selector('v1', 'events')
EVENTS_K8S = Selector('events.k8s.io', 'events')  # only for exclusion from EVERYTHING
NAMESPACES = Selector('v1', 'namespaces')
CLUSTER_PEERINGS = Selector('clusterkopfpeerings')
NAMESPACED_PEERINGS = Selector('kopfpeerings')


class Backbone(Mapping[Selector, Resource]):
    """
    Actual resources used in the core (reactor & engines) of the framework.

    Why? The codebase only refers to the resources by API group/version & names.
    The actual resources can be different in different clusters, usually due
    to different versions: e.g. "v1" vs. "v1beta1" for CRDs.

    The actual backbone resources are detected in the initial cluster scanning
    during the operator startup in :func:`resource_scanner`.

    The backbone resources cannot be changed at runtime after they are found
    for the first time -- since the core tasks are already started with those
    resource definitions, and cannot be easily restarted.

    This does not apply to the resources of the operator (not the framework!),
    where the resources can be created, changed, and deleted at runtime easily.
    """

    def __init__(self) -> None:
        super().__init__()
        self._items: MutableMapping[Selector, Resource] = {}
        self._revised = asyncio.Condition()
        self.selectors = [NAMESPACES, EVENTS, CRDS, CLUSTER_PEERINGS, NAMESPACED_PEERINGS]

    def __len__(self) -> int:
        return len(self._items)

    def __iter__(self) -> Iterator[Selector]:
        return iter(self._items)

    def __getitem__(self, item: Selector) -> Resource:
        return self._items[item]

    async def fill(
            self,
            *,
            resources: Iterable[Resource],
    ) -> None:
        async with self._revised:
            for resource in resources:
                for spec in self.selectors:
                    if spec not in self._items:
                        if spec.check(resource):
                            self._items[spec] = resource
            self._revised.notify_all()

    async def wait_for(
            self,
            selector: Selector,
    ) -> Resource:
        """
        Wait for the actual resource to be found in the cluster scanning.

        The resources can be cached in-memory. Once the resource is retrieved,
        it never changes in memory even if it changes in the cluster. This is
        intentional -- to match with the nature of the cluster scanning,
        which waits for the resources and then starts background jobs,
        which are not easy to terminate without terminating the whole operator.
        """
        async with self._revised:
            await self._revised.wait_for(lambda: selector in self)
        return self[selector]


@dataclasses.dataclass(frozen=True)
class Insights:
    """
    Actual resources & namespaces served by the operator.
    """
    namespaces: Set[Namespace] = dataclasses.field(default_factory=set)
    resources: Set[Resource] = dataclasses.field(default_factory=set)
    backbone: Backbone = dataclasses.field(default_factory=Backbone)

    # Signalled when anything changes in the insights.
    revised: asyncio.Condition = dataclasses.field(default_factory=asyncio.Condition)

    # The flags that are set after the initial listing is finished. Not cleared afterwards.
    ready_namespaces: asyncio.Event = dataclasses.field(default_factory=asyncio.Event)
    ready_resources: asyncio.Event = dataclasses.field(default_factory=asyncio.Event)

    # The resources that are part of indices and can block the operator readiness on start.
    indexable: Set[Resource] = dataclasses.field(default_factory=set)



"""
Testing the handling of events on the top level.

As input:

* Mocked cause detection, with the cause artificially simulated for each test.
  The proper cause detection is tested elsewhere (see ``test_detection.py``).
* Registered handlers in a global registry. Each handler is a normal function,
  which calls a mock -- to ease the assertions.

As output, we check mocked calls on the following:

* ``asyncio.sleep()`` -- for delays.
* ``kopf.clients.patching.patch_obj()`` -- for patch content.
* ``kopf.clients.events.post_event()`` -- for events posted.
* Handler mocks -- whether they were or were not called with specific arguments.
* Captured logs.

The above inputs & outputs represent the expected user scenario
rather than the specific implementation of it.
Therefore, we do not mock/spy/intercept anything within the handling routines
(except for cause detection), leaving it as the implementation details.
Specifically, this internal chain of calls happens on every event:

* ``causation.detect_*_cause()`` -- tested separately in ``/tests/causation/``.
* ``handle_cause()``
* ``execute()``
* ``_execute()``
* ``_call_handler()``
* ``invocation.invoke()`` -- tested separately in ``/tests/invocations/``.

Some of these aspects are tested separately to be sure they indeed execute
all possible cases properly. In the top-level event handling, we assume they do,
and only check for the upper-level behaviour, not all of the input combinations.
"""
import dataclasses
from typing import Callable
from unittest.mock import Mock

import pytest

import kopf
from kopf.reactor.causation import ResourceChangingCause


@dataclasses.dataclass(frozen=True, eq=False)
class K8sMocks:
    patch_obj: Mock
    post_event: Mock
    sleep_or_wait: Mock


@pytest.fixture(autouse=True)
def k8s_mocked(mocker, resp_mocker):
    # We mock on the level of our own K8s API wrappers, not the K8s client.
    return K8sMocks(
        patch_obj=mocker.patch('kopf.clients.patching.patch_obj', return_value={}),
        post_event=mocker.patch('kopf.clients.events.post_event'),
        sleep_or_wait=mocker.patch('kopf.structs.primitives.sleep_or_wait', return_value=None),
    )


@dataclasses.dataclass(frozen=True, eq=False, order=False)
class HandlersContainer:
    index_mock: Mock
    event_mock: Mock
    create_mock: Mock
    update_mock: Mock
    delete_mock: Mock
    resume_mock: Mock
    event_fn: Callable
    create_fn: Callable
    update_fn: Callable
    delete_fn: Callable
    resume_fn: Callable


@pytest.fixture()
def handlers(registry):
    index_mock = Mock(return_value=None)
    event_mock = Mock(return_value=None)
    create_mock = Mock(return_value=None)
    update_mock = Mock(return_value=None)
    delete_mock = Mock(return_value=None)
    resume_mock = Mock(return_value=None)

    @kopf.index('kopfexamples', id='index_fn')
    async def index_fn(**kwargs):
        return index_mock(**kwargs)

    @kopf.on.event('kopfexamples', id='event_fn')
    async def event_fn(**kwargs):
        return event_mock(**kwargs)

    # Keep on-resume on top, to catch any issues with the test design (where it could be skipped).
    @kopf.on.resume('kopfexamples', id='resume_fn', timeout=600, retries=100,
                    deleted=True)  # only for resuming handles, to cover the resource being deleted.
    async def resume_fn(**kwargs):
        return resume_mock(**kwargs)

    @kopf.on.create('kopfexamples', id='create_fn', timeout=600, retries=100)
    async def create_fn(**kwargs):
        return create_mock(**kwargs)

    @kopf.on.update('kopfexamples', id='update_fn', timeout=600, retries=100)
    async def update_fn(**kwargs):
        return update_mock(**kwargs)

    @kopf.on.delete('kopfexamples', id='delete_fn', timeout=600, retries=100)
    async def delete_fn(**kwargs):
        return delete_mock(**kwargs)

    return HandlersContainer(
        index_mock=index_mock,
        event_mock=event_mock,
        create_mock=create_mock,
        update_mock=update_mock,
        delete_mock=delete_mock,
        resume_mock=resume_mock,
        event_fn=event_fn,
        create_fn=create_fn,
        update_fn=update_fn,
        delete_fn=delete_fn,
        resume_fn=resume_fn,
    )


@pytest.fixture()
def extrahandlers(registry, handlers):
    index_mock = Mock(return_value=None)
    event_mock = Mock(return_value=None)
    create_mock = Mock(return_value=None)
    update_mock = Mock(return_value=None)
    delete_mock = Mock(return_value=None)
    resume_mock = Mock(return_value=None)

    @kopf.index('kopfexamples', id='index_fn2')
    async def index_fn2(**kwargs):
        return index_mock(**kwargs)

    @kopf.on.event('kopfexamples', id='event_fn2')
    async def event_fn2(**kwargs):
        return event_mock(**kwargs)

    # Keep on-resume on top, to catch any issues with the test design (where it could be skipped).
    # Note: deleted=True -- only for resuming handles, to cover the resource being deleted.
    @kopf.on.resume('kopfexamples', id='resume_fn2', deleted=True)
    async def resume_fn2(**kwargs):
        return resume_mock(**kwargs)

    @kopf.on.create('kopfexamples', id='create_fn2')
    async def create_fn2(**kwargs):
        return create_mock(**kwargs)

    @kopf.on.update('kopfexamples', id='update_fn2')
    async def update_fn2(**kwargs):
        return update_mock(**kwargs)

    @kopf.on.delete('kopfexamples', id='delete_fn2')
    async def delete_fn2(**kwargs):
        return delete_mock(**kwargs)

    return HandlersContainer(
        index_mock=index_mock,
        event_mock=event_mock,
        create_mock=create_mock,
        update_mock=update_mock,
        delete_mock=delete_mock,
        resume_mock=resume_mock,
        event_fn=event_fn2,
        create_fn=create_fn2,
        update_fn=update_fn2,
        delete_fn=delete_fn2,
        resume_fn=resume_fn2,
    )


@pytest.fixture()
def cause_mock(mocker, settings):
    """
    Mock the resulting _cause_ of the resource change detection logic.

    The change detection is complex, depends on many fields and values, and it
    is difficult to simulate by artificial event bodies, especially its reason.

    Instead, we patch a method which detects the resource changing causes, and
    return a cause with the mocked reason (also, diff, and some other fields).

    The a value of this fixture, a mock is provided with few fields to mock.
    The default is to no mock anything, unless defined in the test, and to use
    the original arguments to the detection method.
    """

    # Use everything from a mock, but use the passed `patch` dict as is.
    # The event handler passes its own accumulator, and checks/applies it later.
    def new_detect_fn(*, finalizer, diff, new, old, **kwargs):

        # For change detection, we ensure that there is no extra cycle of adding a finalizer.
        raw_event = kwargs.pop('raw_event', None)
        raw_body = raw_event['object']
        raw_body.setdefault('metadata', {}).setdefault('finalizers', [finalizer])

        # Pass through kwargs: resource, logger, patch, diff, old, new.
        # I.e. everything except what we mock -- for them, use the mocked values (if not None).
        return ResourceChangingCause(
            reason=mock.reason,
            diff=mock.diff if mock.diff is not None else diff,
            new=mock.new if mock.new is not None else new,
            old=mock.old if mock.old is not None else old,
            **kwargs)

    # Substitute the real cause detector with out own mock-based one.
    mocker.patch('kopf.reactor.causation.detect_resource_changing_cause', new=new_detect_fn)

    # The mock object stores some values later used by the factory substitute.
    # Note: ONLY those fields we mock in the tests. Other kwargs should be passed through.
    mock = mocker.Mock(spec_set=['reason', 'diff', 'new', 'old'])
    mock.reason = None
    mock.diff = None
    mock.new = None
    mock.old = None
    return mock



import asyncio

import async_timeout
import pytest

from kopf.structs.primitives import sleep_or_wait


async def test_the_only_delay_is_awaited(timer):
    async with timer, async_timeout.timeout(1.0) as timeout:
        unslept = await sleep_or_wait(0.10)
    assert 0.10 <= timer.seconds < 0.11
    assert not timeout.expired
    assert unslept is None


async def test_the_shortest_delay_is_awaited(timer):
    async with timer, async_timeout.timeout(1.0) as timeout:
        unslept = await sleep_or_wait([0.10, 0.20])
    assert 0.10 <= timer.seconds < 0.11
    assert not timeout.expired
    assert unslept is None


async def test_specific_delays_only_are_awaited(timer):
    async with timer, async_timeout.timeout(1.0) as timeout:
        unslept = await sleep_or_wait([0.10, None])
    assert 0.10 <= timer.seconds < 0.11
    assert not timeout.expired
    assert unslept is None


@pytest.mark.parametrize('delays', [
    pytest.param([1000, -10], id='mixed-signs'),
    pytest.param([-100, -10], id='all-negative'),
    pytest.param(-10, id='alone'),
])
async def test_negative_delays_skip_sleeping(timer, delays):
    async with timer, async_timeout.timeout(1.0) as timeout:
        unslept = await sleep_or_wait(delays)
    assert timer.seconds < 0.01
    assert not timeout.expired
    assert unslept is None


@pytest.mark.parametrize('delays', [
    pytest.param([], id='empty-list'),
    pytest.param([None], id='list-of-none'),
])
async def test_no_delays_skip_sleeping(timer, delays):
    async with timer, async_timeout.timeout(1.0) as timeout:
        unslept = await sleep_or_wait(delays)
    assert timer.seconds < 0.01
    assert not timeout.expired
    assert unslept is None


async def test_by_event_set_before_time_comes(timer):
    event = asyncio.Event()
    asyncio.get_running_loop().call_later(0.07, event.set)
    async with timer, async_timeout.timeout(1.0) as timeout:
        unslept = await sleep_or_wait(0.10, event)
    assert unslept is not None
    assert 0.02 <= unslept <= 0.04
    assert 0.06 <= timer.seconds <= 0.08
    assert not timeout.expired


async def test_with_zero_time_and_event_initially_cleared(timer):
    event = asyncio.Event()
    event.clear()
    async with timer, async_timeout.timeout(1.0) as timeout:
        unslept = await sleep_or_wait(0, event)
    assert timer.seconds <= 0.01
    assert not timeout.expired
    assert unslept is None


async def test_with_zero_time_and_event_initially_set(timer):
    event = asyncio.Event()
    event.set()
    async with timer, async_timeout.timeout(1.0) as timeout:
        unslept = await sleep_or_wait(0, event)
    assert timer.seconds <= 0.01
    assert not timeout.expired
    assert not unslept  # 0/None; undefined for such case: both goals reached.



import asyncio
import logging

import async_timeout
import pytest

from kopf.clients.watching import streaming_block
from kopf.structs.primitives import ToggleSet


async def test_pausing_is_ignored_if_turned_off(
        resource, namespace, timer, caplog, assert_logs):
    caplog.set_level(logging.DEBUG)

    operator_paused = ToggleSet(any)
    await operator_paused.make_toggle(False)

    async with timer, async_timeout.timeout(0.5) as timeout:
        async with streaming_block(
            resource=resource,
            namespace=namespace,
            operator_paused=operator_paused,
        ):
            pass

    assert not timeout.expired
    assert timer.seconds < 0.2  # no waits, exits as soon as possible
    assert_logs([], prohibited=[
        r"Pausing the watch-stream for",
        r"Resuming the watch-stream for",
    ])


async def test_pausing_waits_forever_if_not_resumed(
        resource, namespace, timer, caplog, assert_logs):
    caplog.set_level(logging.DEBUG)

    operator_paused = ToggleSet(any)
    await operator_paused.make_toggle(True)

    with pytest.raises(asyncio.TimeoutError):
        async with timer, async_timeout.timeout(0.5) as timeout:
            async with streaming_block(
                resource=resource,
                namespace=namespace,
                operator_paused=operator_paused,
            ):
                pass

    assert timeout.expired
    assert timer.seconds >= 0.5
    assert_logs([
        r"Pausing the watch-stream for",
    ], prohibited=[
        r"Resuming the watch-stream for",
    ])


async def test_pausing_waits_until_resumed(
        resource, namespace, timer, caplog, assert_logs):
    caplog.set_level(logging.DEBUG)

    operator_paused = ToggleSet(any)
    conflicts_found = await operator_paused.make_toggle(True)

    async def delayed_resuming(delay: float):
        await asyncio.sleep(delay)
        await conflicts_found.turn_to(False)

    async with timer, async_timeout.timeout(1.0) as timeout:
        asyncio.create_task(delayed_resuming(0.2))
        async with streaming_block(
            resource=resource,
            namespace=namespace,
            operator_paused=operator_paused,
        ):
            pass

    assert not timeout.expired
    assert timer.seconds >= 0.2
    assert timer.seconds <= 0.5
    assert_logs([
        r"Pausing the watch-stream for",
        r"Resuming the watch-stream for",
    ])



"""
Type definitions from optional 3rd-party libraries, e.g. pykube-ng & kubernetes.

This utility does all the trickery needed to import the libraries if possible,
or to skip them and make typing/runtime dummies for the rest of the codebase.
"""
import abc
from typing import Any, Optional


# Since client libraries are optional, support their objects only if they are installed.
# If not installed, use a dummy class to miss all isinstance() checks for that library.
class _dummy: pass


# Do these imports look excessive? ==> https://github.com/python/mypy/issues/10063
# TL;DR: Strictly `from...import...as...`, AND strictly same-named (`X as X`).
try:
    from pykube.objects import APIObject as APIObject
    PykubeObject = APIObject
except ImportError:
    PykubeObject = _dummy

try:
    from kubernetes.client import V1ObjectMeta as V1ObjectMeta, V1OwnerReference as V1OwnerReference
except ImportError:
    V1ObjectMeta = V1OwnerReference = None


# Kubernetes client does not have any common base classes, its code is fully generated.
# Only recognise classes from a specific module. Ignore all API/HTTP/auth-related tools.
class KubernetesModel(abc.ABC):
    @classmethod
    def __subclasshook__(cls, subcls: Any) -> Any:  # suppress types in this hack
        if cls is KubernetesModel:
            if any(C.__module__.startswith('kubernetes.client.models.') for C in subcls.__mro__):
                return True
        return NotImplemented

    @property
    def metadata(self) -> Optional[V1ObjectMeta]:
        raise NotImplementedError

    @metadata.setter
    def metadata(self, _: Optional[V1ObjectMeta]) -> None:
        raise NotImplementedError



"""
Authentication-related structures.

Kopf handles some rudimentary authentication directly, and exposes the ways
to implement custom authentication methods (via `on.login` handlers).

For that, a minimally sufficient data structure is introduced -- both
to bring all the credentials together in a structured and type-annotated way,
and to receive them from the operators' login-handlers with custom auth methods.

The "rudimentary" is defined as the information passed to the HTTP protocol
and TCP/SSL connection only, i.e. everything usable in a generic HTTP client,
and nothing more than that:

* TCP server host & port.
* SSL verification/ignorance flag.
* SSL certificate authority.
* SSL client certificate and its private key.
* HTTP ``Authorization: Basic username:password``.
* HTTP ``Authorization: Bearer token`` (or other schemes: Bearer, Digest, etc).
* URL's default namespace for the cases when this is implied.

.. seealso::
    :func:`authentication` and :mod:`piggybacking`.
"""
import asyncio
import collections
import dataclasses
import random
from typing import AsyncIterable, AsyncIterator, Callable, Dict, List, \
                   Mapping, NewType, Optional, Tuple, TypeVar, cast

from kopf.structs import primitives


class LoginError(Exception):
    """ Raised when the operator cannot login to the API. """


class AccessError(Exception):
    """ Raised when the operator cannot access the cluster API. """


@dataclasses.dataclass(frozen=True)
class ConnectionInfo:
    """
    A single endpoint with specific credentials and connection flags to use.
    """
    server: str  # e.g. "https://localhost:443"
    ca_path: Optional[str] = None
    ca_data: Optional[bytes] = None
    insecure: Optional[bool] = None
    username: Optional[str] = None
    password: Optional[str] = None
    scheme: Optional[str] = None  # RFC-7235/5.1: e.g. Bearer, Basic, Digest, etc.
    token: Optional[str] = None
    certificate_path: Optional[str] = None
    certificate_data: Optional[bytes] = None
    private_key_path: Optional[str] = None
    private_key_data: Optional[bytes] = None
    default_namespace: Optional[str] = None  # used for cluster objects' k8s-events.
    priority: int = 0


_T = TypeVar('_T', bound=object)

# Usually taken from the HandlerId (also a string), but semantically it is on its own.
VaultKey = NewType('VaultKey', str)


@dataclasses.dataclass
class VaultItem:
    """
    The actual item stored in the vault. It is never exposed externally.

    Used for proper garbage collection when the key is removed from the vault
    (to avoid orchestrating extra cache structures and keeping them in sync).

    The caches are populated by `Vault.extended` on-demand.
    """
    info: ConnectionInfo
    caches: Optional[Dict[str, object]] = None


class Vault(AsyncIterable[Tuple[VaultKey, ConnectionInfo]]):
    """
    A store for currently valid authentication methods.

    *Through we call it a vault to add a sense of security.*

    Normally, only one authentication method is used at a time in multiple
    methods and tasks (e.g. resource watching/patching, peering, etc.).

    Multiple methods to represent the same principal is an unusual case,
    but it is also possible as a side effect. Same for multiple distinct
    identities of a single operator.

    The credentials store is created once for an operator (a task),
    and is then used by multiple tasks running in parallel:

    * Consumed by the API client wrappers to authenticate in the API.
    * Reported by the API client wrappers if some of the credentials fail.
    * Populated by the authenticator background task when and if needed.

    .. seealso::
        :func:`reauthenticated_request`/:func:`reauthenticated_stream`
        and :func:`authentication`.
    """
    _current: Dict[VaultKey, VaultItem]
    _invalid: Dict[VaultKey, List[VaultItem]]

    def __init__(
            self,
            __src: Optional[Mapping[str, object]] = None,
    ) -> None:
        super().__init__()
        self._current = {}
        self._invalid = collections.defaultdict(list)
        self._lock = asyncio.Lock()

        if __src is not None:
            self._update_converted(__src)

        # Mark a pre-populated vault to be usable instantly,
        # or trigger the initial authentication for an empty vault.
        self._ready = primitives.Toggle(bool(self))

    def __repr__(self) -> str:
        return f'<{self.__class__.__name__}: {self._current!r}>'

    def __bool__(self) -> bool:
        return bool(self._current)

    async def __aiter__(
            self,
    ) -> AsyncIterator[Tuple[VaultKey, ConnectionInfo]]:
        async for key, item in self._items():
            yield key, item.info

    async def extended(
            self,
            factory: Callable[[ConnectionInfo], _T],
            purpose: Optional[str] = None,
    ) -> AsyncIterator[Tuple[VaultKey, ConnectionInfo, _T]]:
        """
        Iterate the connection info items with their cached object.

        The cached objects are identified by the purpose (an arbitrary string).
        Multiple types of objects can be cached under different names.

        The factory is a one-argument function of a `ConnectionInfo`,
        that returns the object to be cached for this connection info.
        It is called only once per item and purpose.
        """
        purpose = purpose if purpose is not None else repr(factory)
        async for key, item in self._items():
            if item.caches is None:  # quick-check with no locking overhead.
                async with self._lock:
                    if item.caches is None:  # securely synchronised check.
                        item.caches = {}
            if purpose not in item.caches:  # quick-check with no locking overhead.
                async with self._lock:
                    if purpose not in item.caches:  # securely synchronised check.
                        item.caches[purpose] = factory(item.info)
            yield key, item.info, cast(_T, item.caches[purpose])

    async def _items(
            self,
    ) -> AsyncIterator[Tuple[VaultKey, VaultItem]]:
        """
        Yield the raw items as stored in the vault in random order.

        The items are yielded until either all of them are depleted,
        or until the yielded one does not fail (no `.invalidate` call made).
        Restart on every re-authentication (if new items are added).
        """

        # Yield the connection infos until either all of them are depleted,
        # or until the yielded one does not fail (no `.invalidate` call made).
        # Restart on every re-authentication (if new items are added).
        while True:

            # Whether on the 1st run, or during the active re-authentication,
            # ensure that the items are ready before yielding them.
            await self._ready.wait_for(True)

            # Select the items to yield and let it (i.e. a consumer) work.
            async with self._lock:
                yielded_key, yielded_item = self.select()
            yield yielded_key, yielded_item

            # If the yielded item has been invalidated, assume that this item has failed.
            # Otherwise (the item is in the list), it has succeeded -- we are done.
            # Note: checked by identity, in case a similar item is re-added as a different object.
            async with self._lock:
                if yielded_key in self._current and self._current[yielded_key] is yielded_item:
                    break

    def select(self) -> Tuple[VaultKey, VaultItem]:
        """
        Select the next item (not the info!) to try (and do so infinitely).

        .. warning::
            This method is not async/await-safe: if the data change on the go,
            it can lead to improper items returned.
        """
        if not self._current:
            raise LoginError("No valid credentials are available.")
        prioritised: Dict[int, List[Tuple[VaultKey, VaultItem]]]
        prioritised = collections.defaultdict(list)
        for key, item in self._current.items():
            prioritised[item.info.priority].append((key, item))
        top_priority = max(list(prioritised.keys()))
        key, item = random.choice(prioritised[top_priority])
        return key, item

    async def invalidate(
            self,
            key: VaultKey,
            *,
            exc: Optional[Exception] = None,
    ) -> None:
        """
        Exclude the specified credentials, and re-authenticate as needed.

        Multiple calls can be made for a single authenticator and credentials,
        if used for multiple requests at the same time (a common case).
        All of them will be blocked the same way, until one and only one
        re-authentication happens in a background task. They will be
        unblocked at the same instant once the new credentials are ready.

        If the re-authentication fails in the background task, this method
        re-raises the original exception (most likely a HTTP 401 error),
        and lets the client tasks to fail in their own stack.
        The background task continues to run and tries to re-authenticate
        on the next API calls until cancelled due to the operator exit.
        """

        # Exclude the failed connection items from the list of available ones.
        # But keep a short history of invalid items, so that they are not re-added.
        async with self._lock:
            if key in self._current:
                await self._flush_caches(self._current[key])
                self._invalid[key] = self._invalid[key][-2:] + [self._current[key]]
                del self._current[key]
            need_reauth = not self._current  # i.e. nothing is left at all

        # Initiate a re-authentication activity, and block until it is finished.
        if need_reauth:
            await self._ready.turn_to(False)
            await self._ready.wait_for(True)

        # If the re-auth has failed, re-raise the original exception in the current stack.
        # If the original exception is unknown, raise normally on the next iteration's yield.
        # Keep in mind, this routine is called in parallel from many tasks for the same keys.
        async with self._lock:
            if not self._current:
                if exc is not None:
                    raise exc

    async def populate(
            self,
            __src: Mapping[str, object],
    ) -> None:
        """
        Add newly retrieved credentials.

        Used by :func:`authentication` to add newly retrieved credentials
        from the authentication activity handlers. Some of the credentials
        can be duplicates of the existing ones -- only one of them is used then.
        """

        # Remember the new info items (or replace the old ones). If we already see that the item
        # is invalid (as seen in our short per-key history), we keep it as such -- this prevents
        # consistently invalid credentials from causing infinite re-authentication again and again.
        async with self._lock:
            self._update_converted(__src)

        # Notify the consuming tasks (client wrappers) that new credentials are ready to be used.
        # Those tasks can be blocked in `vault.invalidate()` if there are no credentials left.
        await self._ready.turn_to(True)

    async def wait_for_readiness(self) -> None:
        await self._ready.wait_for(True)

    async def wait_for_emptiness(self) -> None:
        await self._ready.wait_for(False)

    async def close(self) -> None:
        """
        Finalize all the cached objects when the operator is ending.
        """
        async with self._lock:
            for key in self._current:
                await self._flush_caches(self._current[key])

    async def _flush_caches(
            self,
            item: VaultItem,
    ) -> None:
        """
        Call the finalizers and garbage-collect the cached objects.

        Mainly used to garbage-collect aiohttp sessions and its derivatives
        when the connection info items are removed from the vault -- so that
        the sessions/connectors would not complain that they were not close.

        Built-in garbage-collection is not sufficient, as it is synchronous,
        and cannot call the async coroutines like `aiohttp.ClientSession.close`.

        .. note::
            Currently, we assume the ``close()`` method only (both sync/async).
            There is no need to generalise to customizable finalizer callbacks.
            This can change in the future.
        """

        # Close the closable objects.
        if item.caches:
            for obj in item.caches.values():
                if hasattr(obj, 'close'):
                    if asyncio.iscoroutinefunction(getattr(obj, 'close')):
                        await getattr(obj, 'close')()
                    else:
                        getattr(obj, 'close')()

        # Garbage-collect other resources (e.g. files, memory, etc).
        item.caches = None

    def _update_converted(
            self,
            __src: Mapping[str, object],
    ) -> None:
        for key, info in __src.items():
            key = VaultKey(str(key))
            if not isinstance(info, ConnectionInfo):
                raise ValueError("Only ConnectionInfo instances are currently accepted.")
            if info not in [data.info for data in self._invalid[key]]:
                self._current[key] = VaultItem(info=info)



import os

import freezegun
import pytest

from kopf.engines.peering import detect_own_id

SAME_GOOD = [
    ('some-host.example.com', 'other-host.example.com'),
    ('some-host', 'other-host'),
]

# Priorities: A (left) should be selected over B (right).
GOOD_BAD = [
    ('some-host.example.com', 'some-host'),
    ('some-host.example.com', '1.2.3.4'),
    ('some-host.example.com', '::1'),
    ('some-host.example.com', '1.0...0.ip6.arpa'),
    ('some-host.example.com', '4.3.2.1.in-addr.arpa'),
    ('some-host.example.com', '1.2.3.4'),
    ('some-host', '1.2.3.4'),
    ('some-host', '::1'),
    ('some-host', '1.0...0.ip6.arpa'),
    ('some-host', '4.3.2.1.in-addr.arpa'),
    ('some-host', '1.2.3.4'),
]


@pytest.fixture(autouse=True)
def _intercept_os_calls(mocker):
    mocker.patch('getpass.getuser', return_value='some-user')
    mocker.patch('socket.gethostname')
    mocker.patch('socket.gethostbyaddr')


@pytest.mark.parametrize('manual', [True, False])
def test_from_a_pod_id(mocker, manual):
    mocker.patch('socket.gethostname', return_value='some-host')
    mocker.patch('socket.gethostbyaddr', side_effect=lambda fqdn: (fqdn, [], []))
    mocker.patch.dict(os.environ, POD_ID='some-pod-1')
    own_id = detect_own_id(manual=manual)
    assert own_id == 'some-pod-1'


def test_suffixes_appended(mocker):
    mocker.patch('random.choices', return_value='random-str')
    mocker.patch('socket.gethostname', return_value='some-host')
    mocker.patch('socket.gethostbyaddr', side_effect=lambda fqdn: (fqdn, [], []))
    with freezegun.freeze_time('2020-12-31T23:59:59.123456'):
        own_id = detect_own_id(manual=False)
    assert own_id == 'some-user@some-host/20201231235959/random-str'


def test_suffixes_ignored(mocker):
    mocker.patch('socket.gethostname', return_value='some-host')
    mocker.patch('socket.gethostbyaddr', side_effect=lambda fqdn: (fqdn, [], []))
    own_id = detect_own_id(manual=True)
    assert own_id == 'some-user@some-host'


@pytest.mark.parametrize('good1, good2', SAME_GOOD)
def test_good_hostnames_over_good_aliases__symmetric(mocker, good1, good2):
    mocker.patch('socket.gethostname', return_value=good1)
    mocker.patch('socket.gethostbyaddr', side_effect=lambda fqdn: (fqdn, [good2], []))
    own_id = detect_own_id(manual=True)
    assert own_id == f'some-user@{good1}'

    mocker.patch('socket.gethostname', return_value=good2)
    mocker.patch('socket.gethostbyaddr', side_effect=lambda fqdn: (fqdn, [good1], []))
    own_id = detect_own_id(manual=True)
    assert own_id == f'some-user@{good2}'


@pytest.mark.parametrize('good1, good2', SAME_GOOD)
def test_good_aliases_over_good_addresses__symmetric(mocker, good1, good2):
    mocker.patch('socket.gethostname', return_value='localhost')
    mocker.patch('socket.gethostbyaddr', side_effect=lambda fqdn: (fqdn, [good1], [good2]))
    own_id = detect_own_id(manual=True)
    assert own_id == f'some-user@{good1}'

    mocker.patch('socket.gethostname', return_value='localhost')
    mocker.patch('socket.gethostbyaddr', side_effect=lambda fqdn: (fqdn, [good2], [good1]))
    own_id = detect_own_id(manual=True)
    assert own_id == f'some-user@{good2}'


@pytest.mark.parametrize('good, bad', GOOD_BAD)
def test_good_aliases_over_bad_hostnames(mocker, good, bad):
    mocker.patch('socket.gethostname', return_value=bad)
    mocker.patch('socket.gethostbyaddr', side_effect=lambda fqdn: (fqdn, [good], []))
    own_id = detect_own_id(manual=True)
    assert own_id == f'some-user@{good}'


@pytest.mark.parametrize('good, bad', GOOD_BAD)
def test_good_addresses_over_bad_aliases(mocker, good, bad):
    mocker.patch('socket.gethostname', return_value='localhost')
    mocker.patch('socket.gethostbyaddr', side_effect=lambda fqdn: (fqdn, [bad], [good]))
    own_id = detect_own_id(manual=True)
    assert own_id == f'some-user@{good}'


@pytest.mark.parametrize('fqdn', [
    'my-host',
    'my-host.local',
    'my-host.localdomain',
    'my-host.local.localdomain',
    'my-host.localdomain.local',
])
def test_useless_suffixes_removed(mocker, fqdn):
    mocker.patch('socket.gethostname', return_value=fqdn)
    mocker.patch('socket.gethostbyaddr', side_effect=lambda fqdn: (fqdn, [], []))
    own_id = detect_own_id(manual=True)
    assert own_id == 'some-user@my-host'
